"""
Module for computing mask defining great circle arc between two endpoints
"""

import warnings
import numpy as np


from .ecco_utils import get_llc_grid
from .llc_array_conversion import llc_tiles_to_xda
from ecco_v4_py import scalar_calc

# -------------------------------------------------------------------------------
# Functions for generating pre-defined section masks
# -------------------------------------------------------------------------------

def get_section_endpoints(section_name):
    """Get the [lon, lat] endpoints associated with a pre-defined section
    e.g.

        >> pt1, pt2 = get_section_endpoints('Drake Passage')
        pt1 = [-68, -54]
        pt2 = [-63, -66]

    These sections mirror the gcmfaces definitions, see
    gcmfaces/gcmfaces_calc/gcmfaces_lines_pairs.m

    Parameters
    ----------
    section_name : str
        name of the section to compute transport across

    Returns
    -------
    pt1, pt2 : array_like
        array with two values, [lon, lat] of each endpoint

    or

    None
        if section_name is not in the pre-defined list of sections
    """

    # Set to input lower case and remove spaces/tabs
    section_name = ''.join(section_name.lower().split())

    # Test to see if name exists in list
    section_list = get_available_sections()
    section_list = [''.join(name.lower().split()) for name in section_list]
    if section_name not in section_list:
        warnings.warn('\nSection name %s unavailable as pre-defined section' % section_name)
        return None

    if section_name == 'drakepassage':
        pt1 = [-68, -54]
        pt2 = [-63, -66]
    elif section_name == 'beringstrait':
        pt1 = [-173, 65.5]
        pt2 = [-164, 65.5]
    elif section_name == 'gibraltar':
        pt1 = [-5, 34]
        pt2 = [-5, 40]
    elif section_name == 'floridastrait':
        pt1 = [-81, 28]
        pt2 = [-77, 26]
    elif section_name == 'floridastraitw1':
        pt1 = [-81, 28]
        pt2 = [-79, 22]
    elif section_name == 'floridastraits1':
        pt1 = [-76, 21]
        pt2 = [-76,  8]
    elif section_name == 'floridastraite1':
        pt1 = [-77, 26]
        pt2 = [-77, 24]
    elif section_name == 'floridastraite2':
        pt1 = [-77, 24]
        pt2 = [-77, 22]
    elif section_name == 'floridastraite3':
        pt1 = [-76, 21]
        pt2 = [-72, 18.5]
    elif section_name == 'floridastraite4':
        pt1 = [-72, 18.5]
        pt2 = [-72, 10]
    elif section_name == 'davisstrait':
        pt1 = [-65, 66]
        pt2 = [-50, 66]
    elif section_name == 'denmarkstrait':
        pt1 = [-35, 67]
        pt2 = [-20, 65]
    elif section_name == 'icelandfaroe':
        pt1 = [-16, 65]
        pt2 = [ -7, 62.5]
    elif section_name == 'faroescotland':
        pt1 = [-6.5, 62.5]
        pt2 = [-4, 57]
    elif section_name == 'scotlandnorway':
        pt1 = [-4, 57]
        pt2 = [ 8, 62]
    elif section_name == 'framstrait':
        pt1 = [-22, 79]
        pt2 = [15, 78.8]
    elif section_name == 'barentssea':
        pt1 = [16.5, 77]
        pt2 = [21.5, 69.6]
    elif section_name == 'labradorgreenland':
        pt1 = [-57, 53]
        pt2 = [-43, 60]
    elif section_name == 'hudsonstrait':
        pt1 = [-65.5, 59]
        pt2 = [-65.5, 63.5]
    elif section_name == 'englishchannel':
        pt1 = [-4.5, 48]
        pt2 = [-5, 50.5]
    elif section_name == 'newfoundlandiberia':
        pt1 = [-53., 47.5]
        pt2 = [-8, 43]
    elif section_name == 'indonesiaw1':
        pt1 = [103, 4]
        pt2 = [103,-1]
    elif section_name == 'indonesiaw2':
        pt1 = [104, -3]
        pt2 = [109, -8]
    elif section_name == 'indonesiaw3':
        pt1 = [113, -8.5]
        pt2 = [118, -8.5]
    elif section_name == 'indonesiaw4':
        pt1 = [118, -8.5]
        pt2 = [127, -15]
    elif section_name == 'australiaantarctica':
        pt1 = [127, -25]
        pt2 = [127, -68]
    elif section_name == 'madagascarchannel':
        pt1 = [38, -10]
        pt2 = [46, -22]
    elif section_name == 'madagascarantarctica':
        pt1 = [46, -22]
        pt2 = [46, -69]
    elif section_name == 'southafricaantarctica':
        pt1 = [20, -30]
        pt2 = [20, -69.5]

    return pt1, pt2

def get_available_sections():
    """Return pre-defined section names for computing transports across this section

    Returns
    -------
    section_list : list of str
        list of available pre-defined sections
    """
    section_list = ['Bering Strait',
    'Gibraltar',
    'Florida Strait',
    'Florida Strait W1',
    'Florida Strait S1',
    'Florida Strait E1',
    'Florida Strait E2',
    'Florida Strait E3',
    'Florida Strait E4',
    'Davis Strait',
    'Denmark Strait',
    'Iceland Faroe',
    'Faroe Scotland',
    'Scotland Norway',
    'Fram Strait',
    'Barents Sea',
    'Labrador Greenland',
    'Hudson Strait',
    'English Channel',
    'Newfoundland Iberia',
    'Drake Passage',
    'Indonesia W1',
    'Indonesia W2',
    'Indonesia W3',
    'Indonesia W4',
    'Australia Antarctica',
    'Madagascar Channel',
    'Madagascar Antarctica',
    'South Africa Antarctica']

    return section_list

# -------------------------------------------------------------------------------
# Main function to compute section masks
# -------------------------------------------------------------------------------

def get_section_line_masks(pt1, pt2, cds, grid=None):
    """Compute 2D mask with 1's along great circle line
    from lat/lon1 -> lat/lon2

    Parameters
    ----------
    pt1, pt2 : tuple or list with 2 floats
        [longitude, latitude] or (longitude, latitude) of endpoints
    cds : xarray Dataset
        containing grid coordinate information, at least XC, YC
    grid : xgcm grid object
        see ecco_utils.get_llc_grid

    Returns
    -------
    section_mask : xarray DataArray
        2D mask along section
    """

    # Get cartesian coordinates of end points
    x1, y1, z1 = _convert_latlon_to_cartesian(pt1[0],pt1[1])
    x2, y2, z2 = _convert_latlon_to_cartesian(pt2[0],pt2[1])

    # Compute rotation matrices
    # 1. Rotate around x-axis to put first point at z = 0
    theta_1 = np.arctan2(-z1, y1)
    rot_1 = np.vstack(( [1, 0, 0],
                        [0, np.cos(theta_1),-np.sin(theta_1)],
                        [0, np.sin(theta_1), np.cos(theta_1)]))

    x1, y1, z1 = _apply_rotation_matrix(rot_1, (x1,y1,z1))
    x2, y2, z2 = _apply_rotation_matrix(rot_1, (x2,y2,z2))

    # 2. Rotate around z-axis to put first point at y = 0
    theta_2 = np.arctan2(x1,y1)
    rot_2 = np.vstack(( [np.cos(theta_2),-np.sin(theta_2), 0],
                        [np.sin(theta_2), np.cos(theta_2), 0],
                        [0, 0, 1]))

    x1, y1, z1 = _apply_rotation_matrix(rot_2, (x1,y1,z1))
    x2, y2, z2 = _apply_rotation_matrix(rot_2, (x2,y2,z2))

    # 3. Rotate around y-axis to put second point at z = 0
    theta_3 = np.arctan2(-z2, -x2)
    rot_3 = np.vstack(( [ np.cos(theta_3), 0, np.sin(theta_3)],
                        [ 0, 1, 0],
                        [-np.sin(theta_3), 0, np.cos(theta_3)]))

    x1, y1, z1 = _apply_rotation_matrix(rot_3, (x1,y1,z1))
    x2, y2, z2 = _apply_rotation_matrix(rot_3, (x2,y2,z2))

    # Now apply rotations to the grid
    # and get cartesian coordinates at cell centers
    xc, yc, zc = _rotate_the_grid(cds.XC, cds.YC, rot_1, rot_2, rot_3)

    # Interpolate for x,y to west and south edges
    grid = get_llc_grid(cds) if grid is None else grid
    xw = grid.interp(xc, 'X', boundary='fill')
    yw = grid.interp(yc, 'X', boundary='fill')
    xs = grid.interp(xc, 'Y', boundary='fill')
    ys = grid.interp(yc, 'Y', boundary='fill')

    # Compute the great circle mask, covering the entire globe
    maskC = scalar_calc.get_edge_mask(zc>0,grid)
    maskW = grid.diff( 1*(zc>0), 'X', boundary='fill')
    maskS = grid.diff( 1*(zc>0), 'Y', boundary='fill')

    # Get section of mask pt1 -> pt2 only
    maskC = _calc_section_along_full_arc_mask(maskC, x1, y1, x2, y2, xc, yc)
    maskW = _calc_section_along_full_arc_mask(maskW, x1, y1, x2, y2, xw, yw)
    maskS = _calc_section_along_full_arc_mask(maskS, x1, y1, x2, y2, xs, ys)

    return maskC, maskW, maskS


# -------------------------------------------------------------------------------
#
# All functions below are non-user facing
#
# -------------------------------------------------------------------------------
# Helper functions for computing section masks
# -------------------------------------------------------------------------------

def _calc_section_along_full_arc_mask( mask, x1, y1, x2, y2, xg, yg ):
    """Given a mask which has a great circle passing through
    pt1 = (x1, y1) and pt2 = (x2,y2), grab the section just connecting pt1 and pt2

    Parameters
    ----------
    mask : xarray DataArray
        2D LLC mask with 1's along great circle across globe, crossing pt1 and pt2
    x1,y1,x2,y2 : scalars
        cartesian coordinates of rotated pt1 and pt2. Note that z1 = z2 = 0
    xg, yg : xarray DataArray
        cartesian coordinates of the rotated horizontal grid

    Returns
    -------
    mask : xarray DataArray
        mask with great arc passing from pt1 -> pt2
    """

    theta_1 = np.arctan2(y1,x1)
    theta_2 = np.arctan2(y2,x2)
    theta_g = np.arctan2(yg,xg)

    if theta_2 < 0:
        theta_g = theta_g.where( theta_g > theta_2, theta_g + 2*np.pi )
        theta_2 = theta_2 + 2 * np.pi

    if (theta_2 - theta_1) <= np.pi:
        mask = mask.where( (theta_g <= theta_2) & (theta_g >= theta_1), 0)
    else:
        mask = mask.where( (theta_g > theta_2) | (theta_g < theta_1), 0)

    return mask

def _rotate_the_grid(lon, lat, rot_1, rot_2, rot_3):
    """Rotate the horizontal grid at lon, lat, via rotation matrices rot_1/2/3

    Parameters
    ----------
    lon, lat : xarray DataArray
        giving longitude, latitude in degrees of LLC horizontal grid
    rot_1, rot_2, rot_3 : np.ndarray
        rotation matrices

    Returns
    -------
    xg, yg, zg : xarray DataArray
        cartesian coordinates of the horizontal grid
    """

    # Get cartesian of 1D view of lat/lon
    lon_v = lon.values.ravel()
    lat_v = lat.values.ravel()
    get_mesh = False
    if not set(('i','j')).issubset(lon.coords):
        lon_v,lat_v = np.meshgrid(lon_v,lat_v)
        lon_v = lon_v.ravel()
        lat_v = lat_v.ravel()
        get_mesh = True
    xg, yg, zg = _convert_latlon_to_cartesian(lon_v,lat_v)

    # These rotations result in:
    #   xg = 0 at pt1
    #   yg = 1 at pt1
    #   zg = 0 at pt1 and pt2 (and the great circle that crosses pt1 & pt2)
    xg, yg, zg = _apply_rotation_matrix(rot_1, (xg,yg,zg))
    xg, yg, zg = _apply_rotation_matrix(rot_2, (xg,yg,zg))
    xg, yg, zg = _apply_rotation_matrix(rot_3, (xg,yg,zg))

    # Remake into LLC xarray DataArray
    if get_mesh:
        xdalike = lat*lon
    else:
        xdalike = lat
    xg = llc_tiles_to_xda(xg, grid_da=xdalike, less_output=True)
    yg = llc_tiles_to_xda(yg, grid_da=xdalike, less_output=True)
    zg = llc_tiles_to_xda(zg, grid_da=xdalike, less_output=True)

    return xg, yg, zg

def _apply_rotation_matrix(rot_mat,xyz):
    """Apply a rotation matrix to a tuple x,y,z (each x,y,z possibly being arrays)

    Parameters
    ----------
    rot_mat : numpy matrix
        2D matrix defining rotation in 3D cartesian coordinates
    xyz : tuple of arrays
        with cartesian coordinates

    Returns
    -------
    xyz_rot : tuple of arrays
        rotated a la rot_mat
    """

    # Put tuple into matrix form
    xyz_mat = np.vstack( (xyz[0],xyz[1],xyz[2]) )

    # Perform rotation
    xyz_rot_mat = np.matmul( rot_mat, xyz_mat )

    # Either return as scalar or array
    if np.isscalar(xyz[0]):
        return xyz_rot_mat[0,0], xyz_rot_mat[1,0], xyz_rot_mat[2,0]
    else:
        return xyz_rot_mat[0,:], xyz_rot_mat[1,:], xyz_rot_mat[2,:]


def _convert_latlon_to_cartesian(lon, lat):
    """Convert latitude, longitude (degrees) to cartesian coordinates
    Note: conversion to cartesian differs from what is found at e.g. Wolfram
    because here lat \in [-pi/2, pi/2] with 0 at equator, not [0, pi], pi/2 at equator

    Parameters
    ----------
    lon : numpy or dask array
        longitude in degrees
    lat : numpy or dask array
        latitude in degrees

    Returns
    -------
    x : numpy or dask array
        x- component of cartesian coordinate
    y : numpy or dask array
    z : numpy or dask array
    """

    # Convert to radians
    lon_r = np.radians(lon)
    lat_r = np.radians(lat)

    # Get cartesian
    x = np.cos(lat_r)*np.cos(lon_r)
    y = np.cos(lat_r)*np.sin(lon_r)
    z = np.sin(lat_r)

    return x, y, z



"""
Compute transport (freshwater, heat, salt) across a section, e.g. Drake Passage
"""

import numpy as np
import xarray as xr
# xarray compatibility
try:
    from xarray.core.pycompat import OrderedDict
except ImportError:
    from collections import OrderedDict

from .ecco_utils import get_llc_grid
from .calc_meridional_trsp import _parse_coords
from .get_section_masks import get_available_sections, \
        get_section_endpoints, get_section_line_masks

# -------------------------------------------------------------------------------
# Main functions for computing standard transport quantities
# -------------------------------------------------------------------------------

# Define constants
METERS_CUBED_TO_SVERDRUPS = 10**-6
WATTS_TO_PETAWATTS = 10**-15
RHO_CONST = 1029
HEAT_CAPACITY = 4000

def calc_section_vol_trsp(ds,
                          pt1=None, pt2=None,
                          section_name=None,
                          maskW=None, maskS=None,
                          coords=None, grid=None):
    """Compute volumetric transport across section in Sverdrups
    There are 3 ways to call this function:

    1. Provide pre-defined section_name, e.g.

        >> trsp = calc_section_vol_trsp(ds,'Drake Passage')

            * Computes volumetric trsp across predefined Drake Passage line
            * See get_available_sections for available definitions

    2. Provide lat/lon pairs to compute transport across, e.g.

        >> pt1 = [lon1, lat1]
        >> pt2 = [lon2, lat2]
        >> trsp = calc_section_vol_trsp(ds,pt1,pt2)

            * Computes volumetric transport across a band between pt1 -> pt2
            * If section name is provided, it gets added to returned DataArray

    3. Provide maskW, maskS, e.g.

        >> _, maskW, maskS = get_section_line_masks(pt1, pt2, ds)
        >> trsp = calc_section_vol_trsp(ds,maskW,maskS)

            * Compute trsp across band defined by masks
            * If section name is provided, it gets added to returned DataArray

    Parameters
    ----------
    ds : xarray Dataset
        must contain UVELMASS,VVELMASS and fields listed under coords below
    pt1, pt2 : list or tuple with two floats, optional
        end points for section line as [lon lat] or (lon, lat)
    maskW, maskS : xarray DataArray, optional
        masks denoting the section, created by get_section_line_masks
    section_name: string, optional
        name for the section. If predefined value, section mask is defined
        via get_section_endpoints
        otherwise, adds name to returned DataArray
    coords : xarray Dataset
        separate dataset containing the coordinate information
        XC, YC, drF, Z, dyG, dxG, optionally maskW, maskS
    grid : xgcm Grid object, optional
        denotes LLC90 operations for xgcm, see ecco_utils.get_llc_grid
        see also the [xgcm documentation](https://xgcm.readthedocs.io/en/latest/grid_topology.html)

    Returns
    -------
    ds_out : xarray Dataset
        includes variables as xarray DataArrays
            vol_trsp
                volumetric transport across section in Sv
                with dimensions 'time' (if in given dataset) and 'lat'
            vol_trsp_z
                volumetric transport across section at each depth level in Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            maskW, maskS
                defining the section
        and the section_name as an attribute if it is provided
    """

    coords = _parse_coords(ds,coords,['Z','YC','XC','drF','dyG','dxG'])

    maskW, maskS = _parse_section_trsp_inputs(coords,pt1,pt2,maskW,maskS,section_name,
                                              grid=grid)

    # Define volumetric transport
    x_vol = ds['UVELMASS'] * coords['drF'] * coords['dyG']
    y_vol = ds['VVELMASS'] * coords['drF'] * coords['dxG']

    # Computes salt transport in m^3/s at each depth level
    ds_out = section_trsp_at_depth(x_vol,y_vol,maskW,maskS,
                                   coords=coords)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'vol_trsp_z'})

    # Sum over depth for total transport
    ds_out['vol_trsp'] = ds_out['vol_trsp_z'].sum('k')

    # Convert both fields to Sv
    for fld in ['vol_trsp','vol_trsp_z']:
        ds_out[fld] = METERS_CUBED_TO_SVERDRUPS * ds_out[fld]
        ds_out[fld].attrs['units'] = 'Sv'

    # Add section name and masks to Dataset
    ds_out['maskW'] = maskW
    ds_out['maskS'] = maskS
    if section_name is not None:
        ds_out.attrs['name'] = section_name

    return ds_out

def calc_section_heat_trsp(ds,
                           pt1=None, pt2=None,
                           section_name=None,
                           maskW=None, maskS=None,
                           coords=None,grid=None):
    """Compute heat transport across section in PW
    Inputs and usage are same as calc_section_vol_trsp.
    The only differences are:

    Parameters
    ----------
    ds : xarray Dataset
        must contain ADVx_TH, ADVy_TH, DFxe_TH, DFyE_TH
    coords : xarray Dataset, optional
        must contain XC, YC, Z, optionally maskW, maskS

    Returns
    -------
    heat_trsp_ds : xarray Dataset
        includes variables as xarray DataArrays
            heat_trsp
                heat transport across section in PW
                with dimensions 'time' (if in given dataset) and 'lat'
            heat_trsp_z
                heat transport across section at each depth level in PW
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            maskW, maskS
                defining the section
        and the section_name as an attribute if it is provided
    """

    coords = _parse_coords(ds,coords,['Z','YC','XC'])

    maskW, maskS = _parse_section_trsp_inputs(coords,pt1,pt2,maskW,maskS,section_name,
                                              grid=grid)

    # Define heat transport
    x_heat = ds['ADVx_TH'] + ds['DFxE_TH']
    y_heat = ds['ADVy_TH'] + ds['DFyE_TH']

    # Computes salt transport in degC * m^3/s at each depth level
    ds_out = section_trsp_at_depth(x_heat,y_heat,maskW,maskS,
                                   coords=coords)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'heat_trsp_z'})

    # Sum over depth for total transport
    ds_out['heat_trsp'] = ds_out['heat_trsp_z'].sum('k')

    # Convert both fields to PW
    for fld in ['heat_trsp','heat_trsp_z']:
        ds_out[fld] = WATTS_TO_PETAWATTS * RHO_CONST * HEAT_CAPACITY * ds_out[fld]
        ds_out[fld].attrs['units'] = 'PW'

    # Add section name and masks to Dataset
    ds_out['maskW'] = maskW
    ds_out['maskS'] = maskS
    if section_name is not None:
        ds_out.attrs['name'] = section_name

    return ds_out

def calc_section_salt_trsp(ds,
                           pt1=None, pt2=None,
                           section_name=None,
                           maskW=None, maskS=None,
                           coords=None, grid=None):
    """Compute salt transport across section in psu*Sv
    Inputs and usage are same as calc_section_vol_trsp.
    The only differences are:

    Parameters
    ----------
    ds : xarray Dataset
        must contain ADVx_SLT, ADVy_SLT, DFxe_SLT, DFyE_SLT
    coords : xarray Dataset, optional
        must contain XC, YC, Z, optionally maskW, maskS

    Returns
    -------
    salt_trsp_ds : xarray Dataset
        includes variables as xarray DataArrays
            salt_trsp
                salt transport across section in psu*Sv
                with dimensions 'time' (if in given dataset) and 'lat'
            salt_trsp_z
                salt transport across section at each depth level in psu*Sv
                with dimensions 'time' (if in given dataset), 'lat', and 'k'
            maskW, maskS
                defining the section
        and the section_name as an attribute if it is provided
    """

    coords = _parse_coords(ds,coords,['Z','YC','XC'])

    maskW, maskS = _parse_section_trsp_inputs(coords,pt1,pt2,maskW,maskS,section_name,
                                              grid=grid)

    # Define salt transport
    x_salt = ds['ADVx_SLT'] + ds['DFxE_SLT']
    y_salt = ds['ADVy_SLT'] + ds['DFyE_SLT']

    # Computes salt transport in psu * m^3/s at each depth level
    ds_out = section_trsp_at_depth(x_salt,y_salt,maskW,maskS,
                                   coords=coords)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'salt_trsp_z'})

    # Sum over depth for total transport
    ds_out['salt_trsp'] = ds_out['salt_trsp_z'].sum('k')

    # Convert both fields to psu.Sv
    for fld in ['salt_trsp','salt_trsp_z']:
        ds_out[fld] = METERS_CUBED_TO_SVERDRUPS * ds_out[fld]
        ds_out[fld].attrs['units'] = 'psu.Sv'

    # Add section name and masks to Dataset
    ds_out['maskW'] = maskW
    ds_out['maskS'] = maskS
    if section_name is not None:
        ds_out.attrs['name'] = section_name

    return ds_out

# -------------------------------------------------------------------------------
# Main function for computing standard transport quantities
# -------------------------------------------------------------------------------

def section_trsp_at_depth(xfld, yfld, maskW, maskS,
                          coords=None):
    """
    Compute transport of vector quantity at each depth level
    across latitude(s) defined in lat_vals

    Parameters
    ----------
    xfld, yfld : xarray DataArray
        3D spatial (+ time, optional) field at west and south grid cell edge
    maskW, maskS : xarray DataArray
        defines the section to define transport across
    coords : xarray Dataset, optional
        include if providing maskW/S (i.e. wet point masks in addition to line masks)

    Returns
    -------
    ds_out : xarray Dataset
        with the main variable
            'trsp_z'
                transport of vector quantity across denoted section at
                each depth level with dimensions 'time' (if in given dataset),
                and 'k' (depth)
    """

    # Initialize empty DataArray with coordinates and dims
    coords = coords if coords is not None else xfld.to_dataset(name='xfld')
    ds_out = _initialize_section_trsp_data_array(coords)

    # Apply section mask and sum horizontally
    # if wet point mask in coords, use it
    maskW = maskW.where(coords['maskW']) if 'maskW' in coords else maskW
    maskS = maskS.where(coords['maskS']) if 'maskS' in coords else maskS
    sec_trsp_x = (xfld * maskW).sum(dim=['i_g','j','tile'])
    sec_trsp_y = (yfld * maskS).sum(dim=['i','j_g','tile'])

    ds_out['trsp_z'] = sec_trsp_x + sec_trsp_y

    return ds_out


# -------------------------------------------------------------------------------
#
# All functions below are non-user facing
#
# -------------------------------------------------------------------------------
# Helper functions for the computing volume, heat, and salt transport
# -------------------------------------------------------------------------------

def _parse_section_trsp_inputs(ds,pt1,pt2,maskW,maskS,section_name,grid=None):
    """Handle inputs for computing volume, heat, or salt transport across
    a section

    Parameters
    ----------
    see calc_section_vol_trsp

    Returns
    -------
    maskW, maskS : xarray DataArray
        masks defining the section
    """

    use_predefined_section = False
    use_endpoints = False
    use_masks = False

    # Test if section name is in available basins
    if section_name is not None:
        if get_section_endpoints(section_name) is not None:
            use_predefined_section = True

    # Test if endpoints provided
    if (pt1 is not None and pt2 is not None):
        use_endpoints = True

    # Test if masks provided
    if (maskW is not None and maskS is not None):
        use_masks = True

    # Test to make sure section is defined by at least one method
    if not use_predefined_section and not use_endpoints and not use_masks:
        raise TypeError('Must provide one method for defining section')

    # First, try to use predefined section
    if use_predefined_section:
        if use_endpoints or use_masks:
            raise TypeError('Cannot provide more than one method for defining section')
        pt1, pt2 = get_section_endpoints(section_name)
    else:
        # Secondly, try to use endpoints or mask
        if use_endpoints and use_masks:
            raise TypeError('Cannot provide more than one method for defining section')
    if not use_masks:
        _, maskW, maskS = get_section_line_masks(pt1, pt2, ds, grid=grid)

    return maskW, maskS

def _initialize_section_trsp_data_array(coords):
    """Create an xarray DataArray with time, depth, and latitude dims

    Parameters
    ----------
    coords : xarray Dataset
        contains LLC coordinates 'k' and (optionally) 'time'

    Returns
    -------
    ds_out : xarray Dataset
        Dataset with the variables
            'trsp_z'
                zero-valued DataArray with time (optional) and
                depth dimensions
            'Z'
                the original depth coordinate
    """

    xda = xr.zeros_like(coords['k'])
    xda = xda if 'time' not in coords.dims else xda.broadcast_like(coords['time']).copy()

    # Convert to dataset to add Z coordinate
    xds = xda.to_dataset(name='trsp_z')
    xds['Z'] = coords['Z']
    xds = xds.set_coords('Z')

    return xds



"""
Some helper functions for plotting
"""

import numpy as np
import xarray as xr

def assign_colormap(fld,user_cmap=None):
    """assign a default colormap based on input field
    following xarray defaults

    Sequential fld: viridis
    Divergent fld : RdBu_r
    

    Parameters
    ----------
    fld : xarray.DataArray or numpy.ndarray
        must be in tiled format

    user_cmap : str or None, optional
        if user_cmap is specified, use this
        None if not specified, and default cmaps selected

    Returns
    -------
    cmap : str
        assigned colormap depending on diverging or sequential
        data in field

    (cmin,cmax) : tuple of floats
        minimum and maximum values for colormap
    """

    #%%    
    # If fld span positive/negative, default to normalize about 0
    # otherwise, regular (sequential). Assign cmap accordingly.
    cmin = np.nanmin(fld)
    cmax = np.nanmax(fld)
    if cmin*cmax<0:
        cmax=np.nanmax(np.abs(fld))
        cmin=-cmax
        cmap = 'RdBu_r'
    else:
        cmap = 'viridis'

    # override if user_cmap is not None
    cmap = cmap if user_cmap is None else user_cmap

    return cmap, (cmin,cmax)



### This function allows users to query ECCO variables and datasets, and then gain access via direct download, or opening files remotely on S3

from .ecco_download import ecco_podaac_query
from .ecco_download import ecco_podaac_download
from .ecco_download import ecco_podaac_download_diskaware
from .ecco_download import ecco_podaac_download
from .ecco_download import ecco_podaac_download_subset

from .ecco_s3_retrieve import ecco_podaac_s3_query
from .ecco_s3_retrieve import ecco_podaac_s3_open
from .ecco_s3_retrieve import ecco_podaac_s3_open_fsspec
from .ecco_s3_retrieve import ecco_podaac_s3_get
from .ecco_s3_retrieve import ecco_podaac_s3_get_diskaware

from .ecco_acc_dates import date_adjustment

from .ecco_varlist import ecco_podaac_varlist_query


import requests
import zarr


def ecco_podaac_access(query,version='v4r4',grid=None,time_res='all',\
                StartDate=None,EndDate=None,snapshot_interval=None,\
                mode='download_ifspace',download_root_dir=None,**kwargs):
    """
    
    This function queries and accesses ECCO datasets from PO.DAAC. The core query and download functions 
    are adapted from Jupyter notebooks created by Jack McNelis and Ian Fenty 
    (https://github.com/ECCO-GROUP/ECCO-ACCESS/blob/master/PODAAC/Downloading_ECCO_datasets_from_PODAAC/README.md)
    and modified by Andrew Delman (https://ecco-v4-python-tutorial.readthedocs.io).
    
    Parameters
    ----------    
    query: str, list, or dict, defines datasets or variables to access.
           If query is str, it specifies either a dataset ShortName (if query 
           matches a NASA Earthdata ShortName), or a text string that can be 
           used to search the ECCO ShortNames, variable names, and descriptions.
           A query may also be a list of multiple ShortNames and/or text searches, 
           or a dict that contains grid,time_res specifiers as keys and ShortNames 
           or text searches as values, e.g.,
           {'native,monthly':['ECCO_L4_SSH_LLC0090GRID_MONTHLY_V4R4',
                              'THETA']}
           will query the native grid monthly SSH datasets, and all native grid 
           monthly datasets with variables or descriptions matching 'THETA'.
    
    version: ('v4r4','v4r5'), specifies ECCO version to query.
             Currently 'v4r5' only works with ['s3_open','s3_get','s3_get_ifspace'] modes, 
             or if the files are already stored in download_root_dir/ShortName/.
             Otherwise an error is returned.
    
    grid: ('native','latlon',None), specifies whether to query datasets with output
          on the native grid or the interpolated lat/lon grid.
          The default None will query both types of grids (and datasets with no spatial
          dimension), unless specified otherwise in a query dict (e.g., the example above).
    
    time_res: ('monthly','daily','snapshot','all'), specifies which time resolution 
              to include in query and downloads. 'all' includes all time resolutions, 
              and datasets that have no time dimension, such as the grid parameter 
              and mixing coefficient datasets.

    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       Full ECCOv4r4 date range (default) is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable 
                       closed budgets within the specified date range.
    
    snapshot_interval: ('monthly', 'daily', or None), if snapshot datasets are included in ShortNames, 
                       this determines whether snapshots are included for only the beginning/end 
                       of each month ('monthly'), or for every day ('daily').
                       If None or not specified, defaults to 'daily' if any daily mean ShortNames 
                       are included and 'monthly' otherwise.
    
    mode: str, one of the following:
          'ls' or 'query': Query dataset ShortNames and variable names/
                           descriptions only; no downloads.
          's3_ls' or 's3_query': Query dataset ShortNames and variable names/
                                 descriptions only; return paths on S3.
          'download': Download datasets using NASA Earthdata URLs
          'download_ifspace': Check storage availability before downloading.
                              Download only if storage footprint of downloads 
                              <= max_avail_frac*(available storage)
          'download_subset': Download spatial and temporal subsets of datasets 
                             via Opendap; query help(ecco_access.ecco_podaac_download_subset)
                             to see keyword arguments that can be used in this mode.
          The following modes work within the AWS cloud only:
          's3_open': Access datasets on S3 without downloading.
          's3_open_fsspec': Use json files (generated with `fsspec` and `kerchunk`) 
                            for expedited opening of datasets.
          's3_get': Download from S3 (to AWS EC2 instance).
          's3_get_ifspace': Check storage availability before downloading; 
                            download if storage footprint 
                            <= max_avail_frac*(available storage).
                            Otherwise data are opened "remotely" from S3 bucket.

    download_root_dir: str, defines parent directory to download files to.
                       Files will be downloaded to directory download_root_dir/ShortName/.
                       If not specified, parent directory defaults to '~/Downloads/ECCO_V4r4_PODAAC/',
                       or '~/Downloads/ECCO_V4r5_PODAAC/' if version == 'v4r5'.
    
    Additional keyword arguments*:
    *This is not an exhaustive list, especially for 
    'download_subset' mode; use help(ecco_access.ecco_podaac_download_subset) to display 
    options specific to that mode
    
    max_avail_frac: float, maximum fraction of remaining available disk space to 
                    use in storing ECCO datasets.
                    If storing the datasets exceeds this fraction, an error is returned.
                    Valid range is [0,0.9]. If number provided is outside this range, it is replaced 
                    by the closer endpoint of the range.
    
    jsons_root_dir: str, for s3_open_fsspec mode only, the root/parent directory where the 
                    fsspec/kerchunk-generated jsons are found.
                    jsons are generated using the steps described here:
                    https://medium.com/pangeo/fake-it-until-you-make-it-reading-goes-netcdf4-data-on-aws-s3
                    as-zarr-for-rapid-data-access-61e33f8fe685.
                    If None (default), jsons_root_dir will be set to ~/MZZ/{version}/.
                    The jsons need to be stored in the following directories/formats:
                    For v4r4: {jsons_root_dir}/MZZ_{GRIDTYPE}_{TIME_RES}/{SHORTNAME}.json.
                    GRIDTYPE is '05DEG' or 'LLC0090GRID' for v4r4;
                    TIME_RES is one of: ('MONTHLY','DAILY','SNAPSHOT','GEOMETRY','MIXING_COEFFS').
                    For v4r5: {jsons_root_dir}/MZZ_{TIME_RES}_{GRIDTYPE}/{SHORTNAME_core}_{TIME_RES}_{GRIDTYPE}_llc090_ECCOV4r5.json.
                    TIME_RES is one of: ('mon_mean','day_mean','snap');
                    GRIDTYPE is one of: ('native','latlon').
                    The SHORTNAME_core is the SHORTNAME without the 'ECCO_L4_' at the beginning, 
                    or the grid/time/version IDs at the end (e.g., 'LLC0090GRID_MONTHLY_V4R5').
    
    jsons_retrieve: bool, if True (default), and jsons are not already stored in jsons_root_dir, 
                    retrieve from S3 bucket and store locally in jsons_root_dir.
                    If False, and jsons are not stored in jsons_root_dir, 
                    the data retrieval will fail.
    
    n_workers: int, number of workers to use in concurrent downloads. Benefits typically taper off above 5-6.
    
    force_redownload: bool, if True, existing files will be redownloaded and replaced;
                            if False (default), existing files will not be replaced.
    
    show_noredownload_msg: bool, if True (default), and force_redownload=False, 
                               display message for each file that is already 
                               downloaded (and therefore not re-downloaded); 
                               if False, these messages are not shown.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.
    
    return_granules: bool, if True (default), str or list of queried or 
                           downloaded granules/files (including ones that 
                           were already on disk and not replaced) is returned.
                           if False, the function returns nothing.

    Returns
    -------
    granule_files: dict with ShortNames as keys; values are URLs or S3 paths
                   (if in 'query' mode), or paths of files that can be 
                   passed directly to xarray (open_dataset or open_mfdataset).
                   Values are of type str if query finds only one granule/file
                   for that ShortName; of type list if query finds 
                   multiple granules in the same dataset; 
                   or of type fsspec.mapping.FSMap if mode = 's3_open_fsspec'.
                   Only returned if return_granules=True (default).
    
    """
    
    pass
    
    
    ## query varlists as needed to obtain shortnames
    
    def shortnames_find(query_list,version,grid,time_res):
        shortnames_list = []
        for query_item in query_list:
            # see if the query is an existing NASA Earthdata ShortName
            # if not, then do a text search of the ECCO variable lists
            response = requests.get(url="https://cmr.earthdata.nasa.gov/search/collections.json", 
                                    params={'ShortName':query_item})
            
            varlist_query = True
            if response.status_code == 200:
                if len(response.json()['feed']['entry']) > 0:
                    # Earthdata CMR found a match for the ShortName; no need to query varlists
                    varlist_query = False
                    shortnames_list.append(query_item)
            else:
                print('NASA Earthdata CMR Search is not working currently.\n'\
                      +'Searching ECCO variable lists...')
            if varlist_query:
                # text search of the ECCO variable lists
                shortname_match = ecco_podaac_varlist_query(query_item,version,grid,time_res)
                shortnames_list.append(shortname_match)
        
        return shortnames_list
    
    
    if isinstance(query,str):
        query = [query]
    if isinstance(query,dict):
        shortnames = []
        for gridtime_spec,curr_query in query.items():
            try:
                curr_grid,curr_time_res = gridtime_spec.split(',')
            except:
                raise ValueError("Keys of dict 'query' must be of the form grid,time_res\n'\
                                 +'with 1 comma in the middle")
            if isinstance(curr_query,str):
                curr_query = [curr_query]
            shortnames += shortnames_find(curr_query,\
                                          version,\
                                          grid=curr_grid,\
                                          time_res=curr_time_res)
    else:
        shortnames = shortnames_find(query,version,grid=grid,time_res=time_res)
    
    
    ## query NASA Earthdata CMR and download granules
    
    possible_mode_list = "['ls','query','s3_ls','s3_query','download',\n"\
                         +"'download_ifspace','download_subset',\n"\
                         +"'s3_open','s3_get','s3_get_ifspace','s3_open_fsspec']"
    
    # set some default keyword arguments
    if (('n_workers' not in kwargs.keys()) and (mode != 'download_subset')):
        kwargs['n_workers'] = 6
    if 'force_redownload' not in kwargs.keys():
        kwargs['force_redownload'] = False

    # remove unneeded keyword arguments
    if mode == 's3_open_fsspec':
        for kwarg in list(kwargs.keys()):
            if kwarg not in ['jsons_root_dir','jsons_retrieve','prompt_request_payer']:
                del kwargs[kwarg]
    else:
        if 'jsons_root_dir' in kwargs.keys():
            del kwargs['jsons_root_dir']
        if 'jsons_retrieve' in kwargs.keys():
            del kwargs['jsons_retrieve']
    if mode == 's3_open':
        for kwarg in list(kwargs.keys()):
            if kwarg in ['n_workers','force_redownload','show_noredownload_msg']:
                del kwargs[kwarg]
    
    
    # download or otherwise access granules, depending on mode
    
    if mode in ['download_ifspace','s3_get_ifspace']:
        if 'max_avail_frac' not in kwargs.keys():
            kwargs['max_avail_frac'] = 0.5
        if mode == 'download_ifspace':
            granule_files = ecco_podaac_download_diskaware(\
                               shortnames,StartDate,EndDate,version,snapshot_interval,\
                               download_root_dir=download_root_dir,**kwargs)
        elif mode == 's3_get_ifspace':
            granule_files = ecco_podaac_s3_get_diskaware(\
                               shortnames,StartDate,EndDate,version,snapshot_interval,\
                               download_root_dir=download_root_dir,**kwargs)
        else:
            raise ValueError('Invalid mode specified; please specify one of the following:'\
              +'\n'+possible_mode_list)
    else:
        if 'max_avail_frac' in kwargs.keys():
            del kwargs['max_avail_frac']
        granule_files = {}
        
        # determine value of snapshot_interval if None or not specified
        if snapshot_interval == None:
            snapshot_interval = 'monthly'
            for curr_shortname in shortnames:
                if 'DAILY' in curr_shortname:
                    snapshot_interval = 'daily'
                    break
        
        for shortname in shortnames:
            
            if mode in ['ls','query']:
                urls,sizes = ecco_podaac_query(shortname,StartDate,EndDate,version,snapshot_interval)
                granule_files[shortname] = urls
            elif mode in ['s3_ls','s3_query']:
                granule_files[shortname] = ecco_podaac_s3_query(\
                                              shortname,StartDate,EndDate,version,snapshot_interval)
            elif mode == 'download':
                kwargs['return_downloaded_files'] = True
                granule_files[shortname] = ecco_podaac_download(\
                                              shortname,StartDate,EndDate,version,snapshot_interval,\
                                              download_root_dir=download_root_dir,\
                                              **kwargs)
            elif mode == 'download_subset':
                if 'n_workers' not in kwargs.keys():
                    kwargs['n_workers'] = 4
                kwargs['return_downloaded_files'] = True
                granule_files[shortname] = ecco_podaac_download_subset(\
                                              shortname,StartDate,EndDate,snapshot_interval,\
                                              download_root_dir=download_root_dir,\
                                              **kwargs)
            elif mode == 's3_open':
                granule_files[shortname] = ecco_podaac_s3_open(\
                                              shortname,StartDate,EndDate,version,snapshot_interval,\
                                              **kwargs)
            elif mode == 's3_open_fsspec':
                # granule_files will consist of mapper objects rather than URL/path or file lists
                granule_files[shortname] = ecco_podaac_s3_open_fsspec(\
                                              shortname,version,**kwargs)
            elif mode == 's3_get':
                kwargs['return_downloaded_files'] = True
                granule_files[shortname] = ecco_podaac_s3_get(\
                                              shortname,StartDate,EndDate,version,snapshot_interval,\
                                              download_root_dir=download_root_dir,\
                                              **kwargs)
            else:
                raise ValueError('Invalid mode specified; please specify one of the following:'\
                  +'\n'+possible_mode_list)
    
    
    # return granule/file list
    
    if 'return_granules' not in kwargs.keys():
        return_granules = True
    if return_granules:
        for shortname in granule_files.keys():
            if isinstance(granule_files[shortname],list):
                if ((len(granule_files[shortname]) == 1) and (mode != 's3_open_fsspec')):
                    # if only 1 file is downloaded, return a string of filename instead of a list
                    granule_files[shortname] = granule_files[shortname][0]
        
        return granule_files



###================================================================================================================



def ecco_podaac_to_xrdataset(query,version='v4r4',grid=None,time_res='all',\
                             StartDate=None,EndDate=None,snapshot_interval=None,\
                             mode='download_ifspace',download_root_dir=None,**kwargs):
    """
    
    This function queries and accesses ECCO datasets from PO.DAAC. The core query and download functions 
    are adapted from Jupyter notebooks created by Jack McNelis and Ian Fenty 
    (https://github.com/ECCO-GROUP/ECCO-ACCESS/blob/master/PODAAC/Downloading_ECCO_datasets_from_PODAAC/README.md)
    and modified by Andrew Delman (https://ecco-v4-python-tutorial.readthedocs.io).
    It is similar to ecco_podaac_access, except instead of a list of URLs or files, 
    an xarray Dataset with all of the queried ECCO datasets is returned.

    Parameters
    ----------    
    query: str, list, or dict, defines datasets or variables to access.
           If query is str, it specifies either a dataset ShortName (if query 
           matches a NASA Earthdata ShortName), or a text string that can be 
           used to search the ECCO ShortNames, variable names, and descriptions.
           A query may also be a list of multiple ShortNames and/or text searches, 
           or a dict that contains grid,time_res specifiers as keys and ShortNames 
           or text searches as values, e.g.,
           {'native,monthly':['ECCO_L4_SSH_LLC0090GRID_MONTHLY_V4R4',
                              'THETA']}
           will query the native grid monthly SSH datasets, and all native grid 
           monthly datasets with variables or descriptions matching 'THETA'.
    
    version: ('v4r4','v4r5'), specifies ECCO version to query.
             Currently 'v4r5' only works with ['s3_open','s3_get','s3_get_ifspace'] modes, 
             or if the files are already stored in download_root_dir/ShortName/.
             Otherwise an error is returned.
    
    grid: ('native','latlon',None), specifies whether to query datasets with output
          on the native grid or the interpolated lat/lon grid.
          The default None will query both types of grids, unless specified 
          otherwise in a query dict (e.g., the example above).
    
    time_res: ('monthly','daily','snapshot','all'), specifies which time resolution 
              to include in query and downloads. 'all' includes all time resolutions, 
              and datasets that have no time dimension, such as the grid parameter 
              and mixing coefficient datasets.

    
    StartDate,EndDate: str, in 'YYYY', 'YYYY-MM', or 'YYYY-MM-DD' format, 
                       define date range [StartDate,EndDate] for download.
                       EndDate is included in the time range (unlike typical Python ranges).
                       Full ECCOv4r4 date range (default) is '1992-01-01' to '2017-12-31'.
                       For 'SNAPSHOT' datasets, an additional day is added to EndDate to enable 
                       closed budgets within the specified date range.
    
    snapshot_interval: ('monthly', 'daily', or None), if snapshot datasets are included in ShortNames, 
                       this determines whether snapshots are included for only the beginning/end 
                       of each month ('monthly'), or for every day ('daily').
                       If None or not specified, defaults to 'daily' if any daily mean ShortNames 
                       are included and 'monthly' otherwise.
    
    mode: str, one of the following:
          'download': Download datasets using NASA Earthdata URLs
          'download_ifspace': Check storage availability before downloading.
                              Download only if storage footprint of downloads 
                              <= max_avail_frac*(available storage)
          'download_subset': Download spatial and temporal subsets of datasets 
                             via Opendap; query help(ecco_access.ecco_podaac_download_subset)
                             to see keyword arguments that can be used in this mode.
          The following modes work within the AWS cloud only:
          's3_open': Access datasets on S3 without downloading.
          's3_open_fsspec': Use json files (generated with `fsspec` and `kerchunk`) 
                            for expedited opening of datasets.
          's3_get': Download from S3 (to AWS EC2 instance).
          's3_get_ifspace': Check storage availability before downloading; 
                            download if storage footprint 
                            <= max_avail_frac*(available storage).
                            Otherwise data are opened "remotely" from S3 bucket.

    download_root_dir: str, defines parent directory to download files to.
                       Files will be downloaded to directory download_root_dir/ShortName/.
                       If not specified, parent directory defaults to '~/Downloads/ECCO_V4r4_PODAAC/',
                       or '~/Downloads/ECCO_V4r5_PODAAC/' if version == 'v4r5'.
    
    Additional keyword arguments*:
    *This is not an exhaustive list, especially for 
    'download_subset' mode; use help(ecco_access.ecco_podaac_download_subset) to display 
    options specific to that mode
    
    max_avail_frac: float, maximum fraction of remaining available disk space to 
                    use in storing ECCO datasets.
                    If storing the datasets exceeds this fraction, an error is returned.
                    Valid range is [0,0.9]. If number provided is outside this range, it is replaced 
                    by the closer endpoint of the range.
    
    jsons_root_dir: str, for s3_open_fsspec mode only, the root/parent directory where the 
                    fsspec/kerchunk-generated jsons are found.
                    jsons are generated using the steps described here:
                    https://medium.com/pangeo/fake-it-until-you-make-it-reading-goes-netcdf4-data-on-aws-s3
                    as-zarr-for-rapid-data-access-61e33f8fe685.
                    If None (default), jsons_root_dir will be set to ~/MZZ/{version}/.
                    The jsons need to be stored in the following directories/formats:
                    For v4r4: {jsons_root_dir}/MZZ_{GRIDTYPE}_{TIME_RES}/{SHORTNAME}.json.
                    GRIDTYPE is '05DEG' or 'LLC0090GRID' for v4r4;
                    TIME_RES is one of: ('MONTHLY','DAILY','SNAPSHOT','GEOMETRY','MIXING_COEFFS').
                    For v4r5: {jsons_root_dir}/MZZ_{TIME_RES}_{GRIDTYPE}/{SHORTNAME_core}_{TIME_RES}_{GRIDTYPE}_llc090_ECCOV4r5.json.
                    TIME_RES is one of: ('mon_mean','day_mean','snap');
                    GRIDTYPE is one of: ('native','latlon').
                    The SHORTNAME_core is the SHORTNAME without the 'ECCO_L4_' at the beginning, 
                    or the grid/time/version IDs at the end (e.g., 'LLC0090GRID_MONTHLY_V4R5').
    
    jsons_retrieve: bool, if True (default), and jsons are not already stored in jsons_root_dir, 
                    retrieve from S3 bucket and store locally in jsons_root_dir.
                    If False, and jsons are not stored in jsons_root_dir, 
                    the data retrieval will fail.
    
    n_workers: int, number of workers to use in concurrent downloads. Benefits typically taper off above 5-6.
    
    force_redownload: bool, if True, existing files will be redownloaded and replaced;
                            if False (default), existing files will not be replaced.
    
    show_noredownload_msg: bool, if True (default), and force_redownload=False, 
                               display message for each file that is already 
                               downloaded (and therefore not re-downloaded); 
                               if False, these messages are not shown.
    
    prompt_request_payer: bool, if True (default), user is prompted to approve 
                                (by entering "y" or "Y") any access to a 
                                requester pays bucket, otherwise request is canceled; 
                                if False, data access proceeds without prompting.

    Returns
    -------
    ds_out: xarray Dataset or dict of xarray Datasets (with ShortNames as keys), 
            containing all of the accessed datasets.
            This function does not work with the query modes: 'ls','query','s3_ls','s3_query'.
    """
    
    pass
    
    
    import numpy as np
    import xarray as xr
    

    # raise error if mode is ls/query only
    if mode in ['ls','query','s3_ls','s3_query']:
        raise ValueError("ecco_podaac_to_xrdataset does not work with 'ls'/'query' modes. \n"\
                         +"Please use ecco_podaac_access with these modes.")
        
        return -1
    
    # submit access query (and download if needed)
    access_output = ecco_podaac_access(query,version,grid,time_res,\
                                       StartDate,EndDate,snapshot_interval,\
                                       mode,download_root_dir,**kwargs)
    
    # determine value of snapshot_interval if None or not specified
    if snapshot_interval == None:
        snapshot_interval = 'monthly'
        for curr_shortname in access_output.keys():
            if 'DAILY' in curr_shortname:
                snapshot_interval = 'daily'
                break
    
    # open xarray datasets
    ds_out = {}
    for shortname,access_out in access_output.items():
        if mode == 's3_open_fsspec':
            curr_ds = xr.open_dataset(access_out,engine='zarr',chunks='auto',consolidated=False)
            if 'time' in curr_ds.dims:
                # isolate time range specified
                startdate,enddate = date_adjustment(shortname,\
                                                    StartDate,EndDate,CMR_query=False)
                time_values = curr_ds.time.values.astype('datetime64[D]')
                in_time_range = np.logical_and(time_values >= startdate,\
                                               time_values <= enddate).nonzero()[0]
                curr_ds = curr_ds.isel(time=in_time_range)
                if (('SNAPSHOT' in shortname) and (snapshot_interval == 'monthly')):
                    month_bounds_list = np.arange(np.datetime64('1992-01','M'),\
                                                  np.datetime64('2040-01','M'),\
                                                  np.timedelta64(1,'M'))\
                                                  .astype('datetime64[D]')
                    time_values = curr_ds.time.values.astype('datetime64[D]')
                    time_subind = list(np.arange(0,len(time_values)).astype('int64'))
                    for count,time_val in enumerate(time_values):
                        if time_val not in month_bounds_list:
                            time_subind.remove(count)
                    curr_ds = curr_ds.isel(time=time_subind)
        else:
            if not isinstance(access_out,list):
                access_out = [access_out]
            curr_ds = xr.open_mfdataset(access_out,\
                                        compat='override',data_vars='minimal',coords='minimal',\
                                        parallel=True)
        ds_out[shortname] = curr_ds
    
    # if only one ShortName is involved, then extract dataset from dictionary
    if len(ds_out) == 1:
        ds_out = list(ds_out.values())[0]
    
    return ds_out



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from __future__ import division,print_function
import numpy as np
import matplotlib.pylab as plt
import xarray as xr
from dask import delayed
import dask

# The Proj class can convert from geographic (longitude,latitude) to native
# map projection (x,y) coordinates and vice versa, or from one map projection
# coordinate system directly to another.
# https://pypi.python.org/pypi/pyproj?
#
import pyresample as pr

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
def resample_to_latlon(orig_lons, orig_lats, orig_field,
                       new_grid_min_lat, 
                       new_grid_max_lat, 
                       new_grid_delta_lat,
                       new_grid_min_lon, 
                       new_grid_max_lon, 
                       new_grid_delta_lon,
                       radius_of_influence = 120000,
                       fill_value = None, mapping_method = 'bin_average') :
    """Take a field from a source grid and interpolate to a target grid.

    Parameters
    ----------
    orig_lons, orig_lats, orig_field : xarray DataArray or numpy array  :
        the lons, lats, and field from the source grid

	new_grid_min_lat, new_grid_max_lat : float
		latitude limits of new lat-lon grid

    new_grid_delta_lat : float
        latitudinal extent of new lat-lon grid cells in degrees (-90..90)

    new_grid_min_lon, new_grid_max_lon : float
		longitude limits of new lat-lon grid (-180..180)

    new_grid_delta_lon : float
         longitudinal extent of new lat-lon grid cells in degrees

    radius_of_influence : float, optional.  Default 120000 m
        the radius of the circle within which the input data is search for
        when mapping to the new grid

    fill_value : float, optional. Default None
		value to use in the new lat-lon grid if there are no valid values
		from the source grid

  	mapping_method : string, optional. Default 'bin_average'
        denote the type of interpolation method to use.
        options include
            'nearest_neighbor' - Take the nearest value from the source grid
            					 to the target grid
            'bin_average'      - Use the average value from the source grid
								 to the target grid

    RETURNS:
    new_grid_lon_centers, new_grid_lat_centers : ndarrays
    	2D arrays with the lon and lat values of the new grid cell centers

    new_grid_lon_edges, new_grid_lat_edges: ndarrays
    	2D arrays with the lon and lat values of the new grid cell edges

    data_latlon_projection:
    	the source field interpolated to the new grid

    """
    if type(orig_lats) == xr.core.dataarray.DataArray:
        orig_lons_1d = orig_lons.values.ravel()
        orig_lats_1d = orig_lats.values.ravel()

    elif type(orig_lats) == np.ndarray:
        orig_lats_1d = orig_lats.ravel()
        orig_lons_1d = orig_lons.ravel()
    else:
        raise TypeError('orig_lons and orig_lats variable either a DataArray or numpy.ndarray. \n'
                'Found type(orig_lons) = %s and type(orig_lats) = %s' %
                (type(orig_lons), type(orig_lats)))
    
    if type(orig_field) == xr.core.dataarray.DataArray:
        orig_field = orig_field.values
    elif type(orig_field) != np.ndarray and \
         type(orig_field) != np.ma.core.MaskedArray :
        raise TypeError('orig_field must be a type of DataArray, ndarray, or MaskedArray. \n'
                'Found type(orig_field) = %s' % type(orig_field))

    ## Modifications to allow time and depth dimensions (DS, 2023-04-20)    
    # Collapse any non-horizontal dimensions into a single, final dimension:

    # Get shape of orig_lats, then difference with orig_field
    n_horiz_dims=len(orig_lats.shape)
    n_total_dims=len(orig_field.shape)
    n_extra_dims=n_total_dims-n_horiz_dims
    horiz_dim_shape=orig_lats.shape # e.g. [13,90,90]    
    if ( (n_extra_dims>0) & (np.prod(orig_field.shape) > np.prod(horiz_dim_shape) ) ):
        # If there are extra dimensions (and they are meaningful/have len > 1)...
        
        # Check if extra dimensions are at beginning or end of orig_field...
        if orig_field.shape[0]!=orig_lats.shape[0]:
            # ... if at the beginning, collapse and move to end
            extra_dims_at_beginning=True
            extra_dim_shape=orig_field.shape[:n_extra_dims] # e.g. [312,50]
            new_shape=np.hstack([np.prod(extra_dim_shape),\
                                 np.prod(horiz_dim_shape)])          # e.g. from [312,50,13,90,90] to [15600,105300]
            orig_field=orig_field.reshape(new_shape).transpose(1,0) # e.g. from [15600,105300] to [105300,15600]
        else:
            # ... if at the end, just collapse
            extra_dims_at_beginning=False
            extra_dim_shape=orig_field.shape[n_horiz_dims:] #e.g. [50,312]
            new_shape=np.hstack([np.prod(horiz_dim_shape),\
                                 np.prod(extra_dim_shape)]) # e.g. from [13,90,90,50,312] to [105300,15600]
            orig_field=orig_field.reshape(new_shape)
    ##


    # prepare for the nearest neighbor mapping

    # first define the lat lon points of the original data
    orig_grid = pr.geometry.SwathDefinition(lons=orig_lons_1d,
                                            lats=orig_lats_1d)


   # the latitudes to which we will we interpolate
    num_lats = int((new_grid_max_lat - new_grid_min_lat) / new_grid_delta_lat + 1)
    num_lons = int((new_grid_max_lon - new_grid_min_lon) / new_grid_delta_lon + 1)

    if (num_lats > 0) and (num_lons > 0):
        # linspace is preferred when using floats!

        new_grid_lat_edges_1D =\
            np.linspace(new_grid_min_lat, new_grid_max_lat, num=int(num_lats))
        
        new_grid_lon_edges_1D =\
            np.linspace(new_grid_min_lon, new_grid_max_lon, num=int(num_lons))

        new_grid_lat_centers_1D = (new_grid_lat_edges_1D[0:-1] + new_grid_lat_edges_1D[1:])/2
        new_grid_lon_centers_1D = (new_grid_lon_edges_1D[0:-1] + new_grid_lon_edges_1D[1:])/2

        new_grid_lon_edges, new_grid_lat_edges =\
            np.meshgrid(new_grid_lon_edges_1D, new_grid_lat_edges_1D)
        
        new_grid_lon_centers, new_grid_lat_centers =\
            np.meshgrid(new_grid_lon_centers_1D, new_grid_lat_centers_1D)

        #print(np.min(new_grid_lon_centers), np.max(new_grid_lon_centers))
        #print(np.min(new_grid_lon_edges), np.max(new_grid_lon_edges))
        
        #print(np.min(new_grid_lat_centers), np.max(new_grid_lat_centers))
        #print(np.min(new_grid_lat_edges), np.max(new_grid_lat_edges)) 
        
        # define the lat lon points of the two parts.
        new_grid  = pr.geometry.GridDefinition(lons=new_grid_lon_centers,
                                               lats=new_grid_lat_centers)

        if mapping_method == 'nearest_neighbor':
            data_latlon_projection = \
                    pr.kd_tree.resample_nearest(orig_grid, orig_field, new_grid,
                                                radius_of_influence=radius_of_influence,
                                                fill_value=fill_value)
        elif mapping_method == 'bin_average':
            wf = lambda r: 1

            data_latlon_projection = \
                    pr.kd_tree.resample_custom(orig_grid, orig_field, new_grid,
                                                radius_of_influence=radius_of_influence,
                                                weight_funcs = wf,
                                                fill_value=fill_value)
        else:
            raise ValueError('mapping_method must be nearest_neighbor or bin_average. \n'
                    'Found mapping_method = %s ' % mapping_method)

        ## Modifications to allow time and depth dimensions (DS, 2023-04-20)
        if ( (n_extra_dims>0) & (np.prod(orig_field.shape) > np.prod(horiz_dim_shape) ) ):
        # If there are extra dimensions (and they are meaningful/have len > 1)
            new_horiz_shape=data_latlon_projection.shape[:2]
            if extra_dims_at_beginning:
                # If the extra dimensions were originally at the beginning, move back...
                data_latlon_projection=data_latlon_projection.transpose(2,0,1)
                # ... and unstack the additional dimensions
                final_shape=np.hstack([extra_dim_shape,new_horiz_shape])
                data_latlon_projection=data_latlon_projection.reshape(final_shape)
            else:
                # If the extra dimensions were originally at the end, just unstack
                final_shape=np.hstack([extra_dim_shape,new_horiz_shape])
                data_latlon_projection=data_latlon_projection.reshape(final_shape)
        ##
        
    else:
        raise ValueError('Number of lat and lon points to interpolate to must be > 0. \n'
                'Found num_lats = %d, num lons = %d' % (num_lats,num_lons))

    return new_grid_lon_centers, new_grid_lat_centers,\
           new_grid_lon_edges, new_grid_lat_edges,\
           data_latlon_projection
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri May 10 17:27:09 2019

@author: ifenty
"""

import sys
import json
import glob
import numpy as np

sys.path.append('/home/ifenty/ECCOv4-py')
import ecco_v4_py as ecco

#%%
######################################################################

## TIME STEPS TO LOAD
time_steps_to_load = [ 732 , 210360]
#time_steps_to_load = 'all'


## GRID DIR -- MAKE SURE USE GRID FIELDS WITHOUT BLANK TILES!!
#mds_grid_dir = '/nobackupp2/ifenty/grids/grid_llc90/no_blank_all'
mds_grid_dir = '/home/ifenty/data/grids/grid_llc90/no_blank_all'

## JSON DIR 
meta_json_dir = '/home/ifenty/ECCOv4-py/meta_json'

## MDS FILE TOP LEVEL DIRECTORY
#mds_diags_root_dir = '/nobackupp7/owang/runs/R3.summerschool/diags'
mds_diags_root_dir = '/home/ifenty/data/model_output/ECCOv4/R3_data_meta_v2/diags_monthly/'

## AVERAGING TIME PERIOD
output_freq_code = 'AVG_MON' ## 'AVG_DAY' or 'SNAPSHOT'

if 'AVG_MON' in output_freq_code:
    files = np.sort(glob.glob(mds_diags_root_dir + '*mon_mean*'))

elif 'AVG_DAY' in output_freq_code:
    files = np.sort(glob.glob(mds_diags_root_dir + '*day_mean*'))

elif 'SNAPSHOT' in output_freq_code:
    files = np.sort(glob.glob(mds_diags_root_dir + '*day_inst*'))
 
## WHICH VARIABLE TO LOAD
vars_to_load = 'all'

## OUTPUT DIRECTORY
output_dir = '/home/ifenty/data/model_output/ECCOv4/netcdf_test2/'

#%%

## START PROGRAM

# load METADATA
# --- common meta data
with open(meta_json_dir + '/ecco_meta_common.json', 'r') as fp:
    meta_common = json.load(fp)

# -- variable specific meta data
with open(meta_json_dir + '/ecco_meta_variable.json', 'r') as fp:    
    meta_variable_specific = json.load(fp)
              
#print (meta_common)
#print (meta_variable_specific)
    
    
## load variable file and directory names
print (len(files))
mds_diagnostic_filesets_to_load = []

for f in files:
    print (f)
    f_basename = str.split(f,'/')[-1]
    
    mds_diagnostic_filesets_to_load.append(f_basename)


print (mds_diagnostic_filesets_to_load)

#%%
## process each file, one at a time.
for mds_diagnostic_fileset_to_load in mds_diagnostic_filesets_to_load:
    
    mds_var_dir = mds_diags_root_dir + '/' + mds_diagnostic_fileset_to_load
    print ('processing ' , mds_diagnostic_fileset_to_load)


    ts = ecco.get_time_steps_from_mds_files(mds_var_dir, mds_diagnostic_fileset_to_load)
    print (mds_var_dir, mds_diagnostic_fileset_to_load)
            
    if 'all' in time_steps_to_load:
        print ('loading all time steps ')
        time_steps_to_load = ts
    
    else:
        print ('\nloading time steps : ', time_steps_to_load)
        
        
    edx, edxp = \
        ecco.create_nc_variable_files_on_native_grid_from_mds(mds_var_dir, 
                         mds_diagnostic_fileset_to_load,
                         mds_grid_dir,
                         output_dir,
                         output_freq_code=output_freq_code,
                         vars_to_load = vars_to_load,
                         tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12],
                         time_steps_to_load = time_steps_to_load,
                         meta_variable_specific = dict(),
                         meta_common = dict(),
                         mds_datatype = '>f4',
                         method = 'time_interval_and_combined_tiles')

    print ('\nfinished processing ' + mds_diagnostic_fileset_to_load  + '\n\n')
   
#



"""
Test routines for the vector calculations module
"""
from __future__ import division, print_function
import warnings
from pathlib import Path
import numpy as np
import xarray as xr
import pytest
from ecco_v4_py import vector_calc, get_llc_grid

from .test_common import all_mds_datadirs, get_test_ds

def test_no_angles(get_test_ds):
    """quick error handling test"""

    ds = get_test_ds
    ds = ds.drop_vars(['CS','SN'])
    with pytest.raises(KeyError):
        vector_calc.UEVNfromUXVY(ds['U'],ds['V'],ds)

def test_optional_grid(get_test_ds):
    """simple, make sure we can optionally provide grid..."""

    ds = get_test_ds
    grid = get_llc_grid(ds)

    uX = xr.ones_like(ds['U'].isel(k=0)).load();
    vY = xr.ones_like(ds['V'].isel(k=0)).load();

    u1,v1 = vector_calc.UEVNfromUXVY(uX,vY,ds)
    u2,v2 = vector_calc.UEVNfromUXVY(uX,vY,ds,grid)
    assert (u1==u2).all()
    assert (v1==v2).all()

def test_uevn_from_uxvy(get_test_ds):
    """make sure grid loc is correct... etc...
    test by feeding right combo of 1, -1's
    so that all velocities are positive 1's ... interp should preserve this
    in the lat/lon components"""

    ds = get_test_ds

    uX,vY = get_fake_vectors(ds['U'],ds['V'])
    uE,vN = vector_calc.UEVNfromUXVY(uX,vY,ds)

    assert set(('i','j')).issubset(uE.dims)
    assert set(('i','j')).issubset(vN.dims)

    # check the lat/lon tiles
    tilelist = [1,4,8,11] if len(uE.tile)==13 else [0,5]
    for t in tilelist:
        assert np.allclose(uE.where(ds.maskC,0.).sel(tile=t),
                           1.*ds.maskC.sel(tile=t),atol=1e-12)
        assert np.allclose(vN.where(ds.maskC,0.).sel(tile=t),
                           1.*ds.maskC.sel(tile=t),atol=1e-12)

def test_latitude_masks(get_test_ds):
    """run through lats, and ensure we're grabbing the closest
    "south-grid-cell-location" (whether that's in the x or y direction!)
    to each latitude value"""

    ds = get_test_ds

    grid = get_llc_grid(ds)
    yW = grid.interp(ds['YC'],'X',boundary='fill')
    yS = grid.interp(ds['YC'],'Y',boundary='fill')
    wetW = ds['maskW'].isel(k=0)
    wetS = ds['maskS'].isel(k=0)

    dLat = 0.5 # is this robust?
    nx = 90

    for lat in np.arange(-89,89,10):
        print('lat: ',lat)
        maskW,maskS = vector_calc.get_latitude_masks(lat,ds['YC'],grid)

        maskW = maskW.where((maskW!=0) & wetW,0.)
        maskS = maskS.where((maskS!=0) & wetS,0.)

        arctic = int((len(maskW.tile)-1)/2)
        assert not (maskW>0).sel(tile=slice(arctic+1,None)).any()
        assert not (maskS<0).sel(tile=slice(arctic-1)).any()

        assert (yW-lat < dLat).where(ds['maskW'].isel(k=0) & (maskW!=0)).all().values
        assert (yS-lat < dLat).where(ds['maskS'].isel(k=0) & (maskS!=0)).all().values

def get_fake_vectors(fldx,fldy):
    fldx.load();
    fldy.load();
    arctic = int((len(fldx.tile)-1)/2)
    for t in fldx.tile.values:
        if t<arctic:
            valx = 1.
            valy = 1.
        elif t == arctic:
            valx = 0.
            valy = 0.
        else:
            valx = -1.
            valy = 1.

        fldx.loc[{'tile':t}] = valx
        fldy.loc[{'tile':t}] = valy
    return fldx, fldy



"""
Test routines for computing meridional transport
"""
import warnings
import numpy as np
import xarray as xr
import pytest
import ecco_v4_py

from .test_common import all_mds_datadirs, get_test_ds
from .test_vector_calc import get_fake_vectors

_section='floridastrait'
_pt1=[-81,28]
_pt2=[-79,22]

@pytest.mark.parametrize("lats",[0,np.array([-20,30,45])])
@pytest.mark.parametrize("basin",[None,'atlExt','pacExt','indExt'])
@pytest.mark.parametrize("doFlip",[True,False])
def test_meridional_stf(get_test_ds,lats,basin,doFlip):
    """compute a meridional streamfunction"""

    ds = get_test_ds
    grid = ecco_v4_py.get_llc_grid(ds)

    ds['U'],ds['V'] = get_fake_vectors(ds['U'],ds['V'])
    ds = ds.rename({'U':'UVELMASS','V':'VVELMASS'})

    if basin is None or len(ds.tile)==13:
        trsp = ecco_v4_py.calc_meridional_stf(ds,lats,doFlip=doFlip,basin_name=basin,grid=grid)

        basinW = ds['maskW']
        basinS = ds['maskS']
        if basin is not None:
            basinW = ecco_v4_py.get_basin_mask(basin,basinW)
            basinS = ecco_v4_py.get_basin_mask(basin,basinS)

        lats = [lats] if np.isscalar(lats) else lats
        for lat in lats:
            maskW,maskS = ecco_v4_py.vector_calc.get_latitude_masks(lat,ds['YC'],grid)

            trspx = (ds['drF']*ds['dyG']*np.abs(maskW)).where(basinW).sum(dim=['i_g','j','tile'])
            trspy = (ds['drF']*ds['dxG']*np.abs(maskS)).where(basinS).sum(dim=['i','j_g','tile'])
            test = trsp.sel(lat=lat).psi_moc.squeeze().reset_coords(drop=True)
            expected = (1e-6*(trspx+trspy)).reset_coords(drop=True)
            if doFlip:
                expected = expected.isel(k=slice(None,None,-1))
            expected=expected.cumsum(dim='k')
            if doFlip:
                expected = -1*expected.isel(k=slice(None,None,-1))
            xr.testing.assert_allclose(test,expected)
    else:
        with pytest.raises(NotImplementedError):
            trsp = ecco_v4_py.calc_meridional_stf(ds,lats,doFlip=doFlip,basin_name=basin,grid=grid)

@pytest.mark.parametrize("args, mask, error",
        [
            ({'section_name':_section,'pt1':None,'pt2':None},False,None),
            ({'section_name':None,    'pt1':_pt1,'pt2':_pt2},False,None),
            ({'section_name':None,    'pt1':None,'pt2':None},True ,None),
            ({'section_name':None,    'pt1':None,'pt2':None},False,TypeError),
            ({'section_name':_section,'pt1':_pt1,'pt2':_pt2},False,TypeError),
            ({'section_name':_section,'pt1':None,'pt2':None},True ,TypeError),
            ({'section_name':"noname",'pt1':None,'pt2':None},False,TypeError),
        ])
@pytest.mark.parametrize("doFlip",[True,False])
def test_section_stf(get_test_ds,args,mask,error,doFlip):
    """compute streamfunction across section"""

    ds = get_test_ds
    grid = ecco_v4_py.get_llc_grid(ds)

    ds['U'],ds['V'] = get_fake_vectors(ds['U'],ds['V'])
    ds = ds.rename({'U':'UVELMASS','V':'VVELMASS'})

    myargs = args.copy()
    if mask:
        myargs['maskW'],myargs['maskS'] = ecco_v4_py.vector_calc.get_latitude_masks(30,ds['YC'],grid)
    else:
        myargs['maskW']=None
        myargs['maskS']=None

    if error is None:
        trsp = ecco_v4_py.calc_section_stf(ds,doFlip=doFlip,grid=grid,**myargs)

        maskW,maskS = ecco_v4_py.calc_section_trsp._parse_section_trsp_inputs(ds,**myargs)

        trspx = (ds['drF']*ds['dyG']*np.abs(maskW)).where(ds['maskW']).sum(dim=['i_g','j','tile'])
        trspy = (ds['drF']*ds['dxG']*np.abs(maskS)).where(ds['maskS']).sum(dim=['i','j_g','tile'])

        test = trsp.psi_moc.squeeze().reset_coords(drop=True)
        expected = (1e-6*(trspx+trspy)).reset_coords(drop=True)
        if doFlip:
            expected = expected.isel(k=slice(None,None,-1))
        expected=expected.cumsum(dim='k')
        if doFlip:
            expected = -1*expected.isel(k=slice(None,None,-1))
        xr.testing.assert_allclose(test,expected)

    else:
        with pytest.raises(error):
            trsp = ecco_v4_py.calc_section_stf(ds,**myargs)

@pytest.mark.parametrize("myfunc, myarg",
        [   (ecco_v4_py.calc_meridional_stf, {'lat_vals':10}),
            (ecco_v4_py.calc_section_stf,{'section_name':_section})])
def test_separate_coords(get_test_ds,myfunc,myarg):
    ds = get_test_ds
    grid = ecco_v4_py.get_llc_grid(ds)

    ds['U'],ds['V'] = get_fake_vectors(ds['U'],ds['V'])
    ds = ds.rename({'U':'UVELMASS','V':'VVELMASS'})

    myarg['grid']=grid
    expected = myfunc(ds,**myarg)
    coords = ds.coords.to_dataset().reset_coords()
    ds = ds.reset_coords(drop=True)

    test = myfunc(ds,coords=coords,**myarg)
    xr.testing.assert_allclose(test['psi_moc'].reset_coords(drop=True),
                               expected['psi_moc'].reset_coords(drop=True))



"""
Test routines for the tile plotting
"""
from __future__ import division, print_function
import warnings
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import pytest
import ecco_v4_py as ecco

from .test_common import llc_mds_datadirs,get_test_array_2d

@pytest.mark.parametrize("vdict",[
        {}, #defaults
        {'cmap':'plasma'},
        {'layout':'latlon',
          'rotate_to_latlon':True},
        {'layout':'latlon',
         'rotate_to_latlon':False,
         'Arctic_cap_tile_location':5},
        {'layout':'latlon',
         'rotate_to_latlon':False,
         'Arctic_cap_tile_location':7},
        {'layout':'latlon',
         'rotate_to_latlon':False,
         'Arctic_cap_tile_location':10},
        {'show_colorbar':True},
        {'show_colorbar':True,
         'show_cbar_label':True,
         'cbar_label':'something good'},
        {'show_tile_labels':False},
        {'fig_size':20},
        {'less_output':False},
        {'show_tile_labels':True,
         'show_colorbar':True},
        {'show_tile_labels':True,
         'show_colorbar':False}
    ])
@pytest.mark.parametrize("is_xda",[True,False])
def test_plot_tiles(get_test_array_2d,vdict,is_xda):
    """Run through various options and make sure nothing is broken"""

    test_arr = get_test_array_2d
    test_arr = test_arr if is_xda else test_arr.values
    ecco.plot_tiles(test_arr,**vdict)
    plt.close()

@pytest.mark.parametrize("vdict",[
        {}, #defaults
        {'cmap':'plasma'},
        {'show_colorbar':True},
        {'show_colorbar':True,
         'show_cbar_label':True,
         'cbar_label':'something good'},
        {'show_tile_labels':False},
        {'less_output':False}
    ])
@pytest.mark.parametrize("is_xda",[True,False])
def test_plot_single_tile(get_test_array_2d,vdict,is_xda):
    """plot a single tile"""

    # read in array
    test_arr = get_test_array_2d
    test_arr = test_arr if is_xda else test_arr.values

    # run plotting routine on each tile
    for t in np.arange(test_arr.shape[0]):
        ecco.plot_tile(test_arr[t,...], **vdict)
        plt.close()


@pytest.mark.parametrize("is_xda",[True,False])
def test_plot_tiles_array(get_test_array_2d,is_xda):
    """a crude test to make sure the array being created
    matches the original"""

    arr_expected = get_test_array_2d
    arr_expected = arr_expected if is_xda else arr_expected.values
    _,arr_test = ecco.plot_tiles(arr_expected)
    assert np.allclose(np.nansum(arr_test),float(np.nansum(arr_expected)))
    plt.close()



"""
Helper functions for all tests
"""
import pytest
from xmitgcm.utils import get_extra_metadata
from xmitgcm.test.test_xmitgcm_common import (
        _experiments, llc_mds_datadirs, setup_mds_dir, dlroot )
import ecco_v4_py as ecco

# Following xmitgcm's lead to add an ASTE domain for testing
_experiments['aste270']= {'geometry':'llc',
                          'dlink': dlroot + '25286756',
                          'md5': 'f616fe46330f1125472f274af2c96e44',
                          'shape': (50,6,270,270),
                          'ref_date':'2002-01-01 00:00:00',
                          'diagnostics':('state_2d_set1',
                            ['ETAN    ', 'SIarea  ', 'SIheff  ', 'SIhsnow ',
                            'DETADT2 ', 'PHIBOT  ', 'sIceLoad', 'MXLDEPTH',
                            'oceSPDep', 'SIatmQnt', 'SIatmFW ', 'oceQnet ',
                            'oceFWflx', 'oceTAUX ', 'oceTAUY ', 'ADVxHEFF',
                            'ADVyHEFF', 'DFxEHEFF', 'DFyEHEFF', 'ADVxSNOW',
                            'ADVySNOW', 'DFxESNOW', 'DFyESNOW', 'SIuice  ',
                            'SIvice  ' 'ETANSQ  ']),
                          'test_iternum':8}

# global variables that will be filled with the directory information
# about the mds and llc files after downloading (setup_mds_dir 
# and llc_mds_datadirs). Once we have that directory information,
# we do not have to redownload the files. This approach significantly
# reduces the local disk space requirement for performing the tests
# using the pytest-xdist parallel package
# -- initially defined as 'None' = not yet filled
mds_dir_info = None
llc_dir_info = None

# Modify xmitgcm's function for both global ECCO and ASTE
@pytest.fixture(scope='module', params=['global_oce_llc90','aste270'])
def all_mds_datadirs(tmpdir_factory, request):

    global mds_dir_info
    if mds_dir_info == None:
    #save the directory information in the global variable
        mds_dir_info = setup_mds_dir(tmpdir_factory,request, _experiments)
   
    print('-------- made mds_dirr_info ')
    print(mds_dir_info) 
    return mds_dir_info
    

@pytest.fixture(scope='module')
def get_test_ds(all_mds_datadirs):
    """make 2 tests when called, one with global, one with ASTE,
    using fixture above"""

    dirname, expected = all_mds_datadirs

    kwargs = {}
    if 'aste' in dirname:
        kwargs['extra_metadata']=get_extra_metadata('aste',270)
        kwargs['tiles_to_load']=[0,1,2,3,4,5]
        kwargs['nx']=270
        domain = 'aste'
    else:
        domain = 'global'

    # read in array
    ds = ecco.load_ecco_vars_from_mds(dirname,
            model_time_steps_to_load=expected['test_iternum'],
            mds_files=['state_2d_set1','U','V','W','T','S'],
            **kwargs)
    ds.load()
    ds.attrs['domain'] = domain
    return ds

@pytest.fixture(scope='module')
def get_global_ds(llc_mds_datadirs):
    """just get the global dataset"""

    global llc_dir_info
    if llc_dir_info == None:
        #save the directory information in the global variable
        dirname, expected = llc_mds_datadirs
        llc_dir_info = [dirname, expected]
    else:
        dirname = llc_dir_info[0]
        expected = llc_dir_info[1] 

    # read in array
    ds = ecco.load_ecco_vars_from_mds(dirname,
            model_time_steps_to_load=expected['test_iternum'],
            mds_files=['state_2d_set1','U','V','W','T','S'])
    ds.load()
    return ds

@pytest.fixture(scope='module')
def get_test_array_2d(llc_mds_datadirs):
    """download, unzip and return 2D field"""
    global llc_dir_info
    if llc_dir_info == None:
        #save the directory information in the global variable
        dirname, expected = llc_mds_datadirs
        llc_dir_info = [dirname, expected]
    else:
        dirname = llc_dir_info[0]
        expected = llc_dir_info[1] 
    
    # read in array
    ds = ecco.load_ecco_vars_from_mds(dirname,
            model_time_steps_to_load=expected['test_iternum'],
            mds_files='state_2d_set1')

    ds.load()
    xda = ds['ETAN']

    if 'time' in xda.dims:
        xda = xda.isel(time=0)
    return xda




import warnings
from datetime import datetime
import numpy as np
import xarray as xr
import pytest
import ecco_v4_py

from .test_common import all_mds_datadirs, get_test_ds

@pytest.mark.parametrize("mytype",['xda','nparr','list','single'])
def test_extract_dates(mytype):

    dints = [[1991,8,9,13,10,15],[1992,10,20,8,30,5]]


    dates = [datetime(year=x[0],month=x[1],day=x[2],
                      hour=x[3],minute=x[4],second=x[5]) for x in dints]
    dates = np.array(dates,dtype='datetime64[s]')
    dates = [np.datetime64(x) for x in dates]
    if mytype=='xda':
        dates = xr.DataArray(np.array(dates))
    elif mytype=='nparr':
        dates = np.array(dates)
    elif mytype=='single':
        dints=dints[0]
        dates = dates[0]

    test_out = ecco_v4_py.extract_yyyy_mm_dd_hh_mm_ss_from_datetime64(dates)
    for test,expected in zip(test_out,np.array(dints).T):
        print('test: ',test)
        print('exp: ',expected)
        test = test.values if mytype=='xda' else test

        assert np.all(test==expected)

def test_get_grid(get_test_ds):
    """make sure we can make a grid ... that's it"""

    grid = ecco_v4_py.get_llc_grid(get_test_ds)



"""
ECCO v4 Python: tile_plot_proj

This module includes routines for plotting arrays in different
projections.

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""

from __future__ import division,print_function
import numpy as np
import matplotlib.pylab as plt
import matplotlib.path as mpath
import cartopy.crs as ccrs
import cartopy as cartopy
import cartopy.feature as cfeature
from .resample_to_latlon import resample_to_latlon

from .plot_utils import assign_colormap

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
def plot_proj_to_latlon_grid(lons, lats, data,
                             projection_type = 'robin',
                             dx=.25, dy=.25,
                             mapping_method='nearest_neighbor',
                             radius_of_influence = 112000,
                             plot_type = 'pcolormesh',
                             circle_boundary = False,
                             cmap = None,
                             cmin = None, 
                             cmax = None,
                             user_lon_0 = 0,
                             user_lat_0 = None,
                             lat_lim = 50,
                             parallels = None,
                             show_coastline = True,
                             show_colorbar = False,
                             show_land = True,
                             show_grid_lines = True,
                             show_grid_labels = False,
                             show_coastline_over_data = True,
                             show_land_over_data = True,
                             grid_linewidth = 1,
                             grid_linestyle = '--',
                             colorbar_label = None,
                             colorbar_location = None,
                             subplot_grid=None,
                             less_output=True,
                             **kwargs):
    
    """Plot a field of data from an arbitrary projection with lat/lon coordinates
    on a geographic projection after resampling it to a regular lat/lon grid.
    

    Parameters
    ----------
    lons, lats : numpy ndarray or xarray DataArrays, required
        the longitudes and latitudes of the data to plot
        
    data : numpy ndarray or xarray DataArray, required
        the field to be plotted
        
    dx, dy : float, optional, default 0.25 degrees
        latitude, longitude spacing of the new lat/lon grid onto which the 
        field 'data' will be resampled.
        
  	mapping_method : string, optional. Default 'nearest_neighbor'
        denote the type of interpolation method to use.
        options include
            'nearest_neighbor' - Take the nearest value from the source grid
            					 to the target grid
            'bin_average'      - Use the average value from the source grid
								 to the target grid        

    radius_of_influence : float, optional, default 112000 m
        to map values from 'data' to the new lat/lon grid, we use use a
        nearest neighbor approach with the constraint that we only use values
        from 'data' that fall within a circle with radius='radius_of_influence'
        from the center of each new lat/lon grid cell. 
        for the llc90, with 1 degree resolution, 
            radius_of_influence = 1/2 x sqrt(2) x 112e3 km 
        would suffice.
        
    projection_type : string, optional
        denote the type of projection, see Cartopy docs.
        options include
            'robin' - Robinson
            'PlateCarree' - flat 2D projection
            'LambertConformal'
            'Mercator'
            'EqualEarth'
            'Mollweide'
            'AlbersEqualArea'
            'cyl' - Lambert Cylindrical
            'ortho' - Orthographic
            'stereo' - polar stereographic projection, see lat_lim for choosing
            'InterruptedGoodeHomolosine'
                North or South
        
    plot_type : string, optional
        denotes type of plot ot make with the data
        options include
            'pcolormesh' - pcolormesh
            'contourf' - filled contour
            'points' - plot points at lat/lon locations

    circle_boundary : logical, optional, default False
        use a circle boundary or not 
    
    cmap : matplotlib.colors.Colormap, optional, default None
        a colormap for the figure.  
    
    cmin/cmax : floats, optional, default None
        the minimum and maximum values to use for the colormap
        if not specified, use the full range of the data
            
    user_lon_0 : float, optional, default 0 degrees
        denote central longitude

    user_lat_0 : float, optional, default None
        denote central latitude (for relevant projections only, see Cartopy)

    lat_lim : int, optional, default 50 degrees
        for stereographic projection, denote the Southern (Northern) bounds for
        North (South) polar projection or cutoff for LambertConformal projection

    parallels : float, optional,
        standard_parallels, one or two latitudes of correct scale
        (for relevant projections only, see Cartopy docs)

    show_coastline : logical, optional, default True
        show coastline or not
                             
    show_colorbar : logical, optional, default False
        show a colorbar or not,

    show_land : logical, optional, default True
        show land or not

    show_grid_lines : logical, optional, default True
        True only possible for some cartopy projections

    show_grid_labels: logical, optional, default False
        True only possible for some cartopy projections

    show_coastline_over_data : logical, optional, default True
        draw coastline over the data or under the data

    show_land_over_data: logical, optional, default True
        draw land over the data or under the data

    grid_linewidth : float, optional, default 1.0
        width of grid lines

    grid_linestyle : string, optional, default = '--'
        pattern of grid lines,
                
    subplot_grid : dict or list, optional
        specifying placement on subplot as
            dict:
                {'nrows': rows_val, 'ncols': cols_val, 'index': index_val}

            list:
                [nrows_val, ncols_val, index_val]

            equates to

                matplotlib.pyplot.subplot(
                    row=nrows_val, col=ncols_val,index=index_val)

    less_output : string, optional
        debugging flag, don't print if True

    """

    if cmap == None:
        cmap, (new_cmin,new_cmax) = assign_colormap(data, cmap)
    
    if cmin == None:
        cmin = np.nanmin(data[:])
    if cmax == None:
        cmax = np.nanmax(data[:])

    if projection_type == 'stereo' and user_lat_0 == None:
        if lat_lim > 0:
            user_lat_0 = 90
        else:
            user_lat_0 = -90

    # Make projection axis
    ax  = _create_projection_axis(
            projection_type, user_lon_0, user_lat_0, parallels,
            lat_lim, subplot_grid, less_output)
    
    ax.set_global()

    # lat-lon data is EPSG:4326
    # https://spatialreference.org/ref/epsg/wgs-84/
    # +proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs 
    data_epsg_code = 4326 
    
    # get current figure
    f = plt.gcf()
    
    # initialize some variables that may not get defined if there is an error
    p = []
    cbar = []
    gl = []
    
    # Polar Stereographic
    if isinstance(ax.projection, ccrs.NorthPolarStereo) or \
       isinstance(ax.projection, ccrs.SouthPolarStereo) :
          
        new_grid_lon_centers, new_grid_lat_centers,\
        new_grid_lon_edges, new_grid_lat_edges,\
        data_latlon_projection = \
            resample_to_latlon(lons, lats, data,
                               -90, 90, dy,
                               -180, 180, dx,
                               mapping_method=mapping_method,
                               radius_of_influence = radius_of_influence)
            
        if plot_type == 'pcolormesh':
            plot_lons = new_grid_lon_edges
            plot_lats = new_grid_lat_edges
        else:
            plot_lons = new_grid_lon_centers
            plot_lats = new_grid_lat_centers
            
        p, gl, cbar = \
            plot_pstereo(plot_lons, 
                         plot_lats, 
                         data_latlon_projection,
                         data_epsg_code,
                         lat_lim,
                         cmin, cmax, ax,
                         plot_type = plot_type,
                         circle_boundary = circle_boundary,
                         cmap = cmap,
                         show_coastline = show_coastline,
                         show_colorbar = show_colorbar,
                         show_land = show_land,
                         show_grid_lines = show_grid_lines,
                         show_grid_labels = show_grid_labels,
                         show_coastline_over_data = show_coastline_over_data,
                         show_land_over_data = show_land_over_data,
                         grid_linewidth = grid_linewidth,
                         grid_linestyle = grid_linestyle,
                         colorbar_label = colorbar_label,
                         colorbar_location = colorbar_location,
                         less_output=less_output)

        new_grid_lon_centers_out = new_grid_lon_centers
        new_grid_lat_centers_out = new_grid_lat_centers
        data_latlon_projection_out = data_latlon_projection

    else: # not polar stereographic projection
    
        # To avoid plotting problems around the date line, lon=180E, -180W
        # we may have to break the plotting into two parts
        # also, for some reason cartopy or matplotlib 
        # doesn't like it when the edge of the 
        # longitude array is the same as the edge of the map when 
        # user_lon_0 (center_longitude) is not 0 or 180/-180.  A small 
        # the longitude grid is therefore adjusted by an epsilon.
        ep = 0.00001
        
        lon_tmp_d = []
        if np.abs(user_lon_0 ) == 180:
            lon_tmp_d.append([ep, 180])
            lon_tmp_d.append([-180, -ep])
        elif user_lon_0 < 0:
            lon_tmp_d.append([-180,180+user_lon_0])
            lon_tmp_d.append([180 + user_lon_0 + ep, 180])
        elif user_lon_0 > 0:
            lon_tmp_d.append([-180+user_lon_0, 180])
            lon_tmp_d.append([-180, -180+user_lon_0-ep])
        elif user_lon_0 == 0:
            # if the map is centered exactly at 0E 
            # cartopy has no problem including -180 and 180 in the longitude
            # arrays
            lon_tmp_d.append([-180, 180])
            
        # loop through different parts of the map to plot (if they exist),
        # do interpolation and plot

        for ki, k in enumerate(lon_tmp_d):
            lon_limits = lon_tmp_d[ki]
            
            new_grid_lon_centers, new_grid_lat_centers,\
            new_grid_lon_edges, new_grid_lat_edges,\
            data_latlon_projection = \
                resample_to_latlon(lons, lats, data,
                                   -90, 90, dy,
                                   lon_limits[0], lon_limits[1], dx,
                                   mapping_method=mapping_method,
                                   radius_of_influence = radius_of_influence)
            if(ki==0):
                new_grid_lon_centers_out = new_grid_lon_centers
                new_grid_lat_centers_out = new_grid_lat_centers
                data_latlon_projection_out = data_latlon_projection
            else:
                new_grid_lon_centers_out = np.append(new_grid_lon_centers_out, new_grid_lon_centers,axis=1)
                new_grid_lat_centers_out = np.append(new_grid_lat_centers_out, new_grid_lat_centers,axis=1)
                data_latlon_projection_out = np.append(data_latlon_projection_out, data_latlon_projection,axis=1)
            if plot_type == 'pcolormesh':
                plot_lons = new_grid_lon_edges
                plot_lats = new_grid_lat_edges
            else:
                plot_lons = new_grid_lon_centers
                plot_lats = new_grid_lat_centers
                
            # don't make the colorbar a second time if we've already
            # made it once in the loop
            if cbar != []:
                show_colorbar=False;

            p, gl, cbar = \
                  plot_global(plot_lons,
                              plot_lats,
                              data_latlon_projection,
                              data_epsg_code,
                              cmin, cmax, ax,
                              plot_type = plot_type,
                              circle_boundary = circle_boundary,
                              cmap = cmap,
                              show_coastline = show_coastline,
                              show_colorbar = show_colorbar,
                              show_land = show_land,
                              show_grid_lines = show_grid_lines,
                              show_grid_labels = show_grid_labels,
                              show_coastline_over_data = show_coastline_over_data,
                              show_land_over_data = show_land_over_data,
                              grid_linewidth = grid_linewidth,
                              grid_linestyle = grid_linestyle,
                              colorbar_label = colorbar_label,
                              colorbar_location = colorbar_location,
                              less_output=less_output)
    

    return f, ax, p, cbar, new_grid_lon_centers_out, new_grid_lat_centers_out,\
        data_latlon_projection_out, gl
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


def plot_pstereo(xx, yy, data,
                 data_epsg_code, \
                 lat_lim,
                 cmin, cmax, ax,
                 plot_type = 'pcolormesh',
                 circle_boundary = False,
                 cmap = None,
                 show_coastline = True,
                 show_colorbar = False,
                 show_land = True,
                 show_grid_lines = True,
                 show_grid_labels = False,
                 show_coastline_over_data = True,
                 show_land_over_data = True,
                 grid_linewidth = 1,
                 grid_linestyle = '--',
                 levels = 20,
                 data_zorder = 50,
                 points_color = 'k',
                 colorbar_label = None,
                 colorbar_location = None,
                 less_output = True):

    # assign cmap default
    if cmap is None:
        cmap, (new_cmin,new_cmax) = assign_colormap(data, cmap)

    if isinstance(ax.projection, ccrs.NorthPolarStereo):
        ax.set_extent([-180, 180, lat_lim, 90], ccrs.PlateCarree())
        if not less_output:
            print('North Polar Projection')
    
    elif isinstance(ax.projection, ccrs.SouthPolarStereo):
        ax.set_extent([-180, 180, -90, lat_lim], ccrs.PlateCarree())
        if not less_output:
            print('South Polar Projection')
    else:
        raise ValueError('ax must be either ccrs.NorthPolarStereo or ccrs.SouthPolarStereo')

    if not less_output:
        print('lat_lim: ',lat_lim)

    # truncate the plot to just a circular shape
    if circle_boundary:
        theta = np.linspace(0, 2*np.pi, 100)
        center, radius = [0.5, 0.5], 0.5
        verts = np.vstack([np.sin(theta), np.cos(theta)]).T
        circle = mpath.Path(verts * radius + center)
        ax.set_boundary(circle, transform=ax.transAxes)

    if data_epsg_code == 4326: # lat lon does nneed to be projected
        data_crs =  ccrs.PlateCarree()
    else:
        # reproject the data if necessary
        data_crs=ccrs.epsg(data_epsg_code)


    # plot the data on zorder 50
    p=[]
    if plot_type == 'pcolormesh':
        p = ax.pcolormesh(xx, yy, data, transform=data_crs,
                          vmin=cmin, vmax=cmax, cmap=cmap, zorder=data_zorder)

    elif plot_type =='contourf':
        p = ax.contourf(xx, yy, data, levels, transform=data_crs,
                        vmin=cmin, vmax=cmax, cmap=cmap, zorder=data_zorder)

    elif plot_type == 'points':
        p = ax.plot(xx, yy,  color=points_color, marker='.', transform=data_crs,
                    zorder=data_zorder)

    else:
        raise ValueError('plot_type  must be either "pcolormesh", "contourf", or "points"')


    gl, cbar =\
        _add_features_to_axis(ax, p, cmin, cmax,
                              cmap = cmap,
                              show_coastline = show_coastline,
                              show_colorbar = show_colorbar,
                              show_land = show_land,
                              show_grid_lines = show_grid_lines,
                              show_grid_labels = show_grid_labels,
                              show_coastline_over_data = show_coastline_over_data,
                              show_land_over_data = show_land_over_data,
                              grid_linewidth = grid_linewidth,
                              grid_linestyle = grid_linestyle,
                              colorbar_label = colorbar_label,
                              colorbar_location = colorbar_location)
        

    return p, gl, cbar

#%%

def plot_global(xx,yy, data,
                data_epsg_code,
                cmin, cmax, ax,
                cmap=None,
                plot_type = 'pcolormesh',
                circle_boundary = False,
                show_coastline = True,
                show_colorbar = False,
                show_land = True,
                show_grid_lines = True,
                show_grid_labels = False,
                show_coastline_over_data = True,
                show_land_over_data = True,
                grid_linewidth = 1,
                grid_linestyle = '--',
                levels = 20,
                colorbar_label=None,
                colorbar_location=None,
                data_zorder = 50,
                points_color = 'k',
                less_output=True):

    # assign cmap default
    if cmap is None:
        cmap, (new_cmin,new_cmax) = assign_colormap(data, cmap)

    if data_epsg_code == 4326: # lat lon does nneed to be projected
        data_crs =  ccrs.PlateCarree()
    else:
        data_crs =ccrs.epsg(data_epsg_code)

    # plot the data on zorder 50
    p=[]
    if plot_type == 'pcolormesh':
        p = ax.pcolormesh(xx, yy, data, transform=data_crs, \
                          vmin=cmin, vmax=cmax, cmap=cmap,
                          zorder=data_zorder)

    elif plot_type =='contourf':
        p = ax.contourf(xx, yy, data, levels, transform=data_crs,  \
                        vmin=cmin, vmax=cmax, cmap=cmap,
                        zorder=data_zorder)

    elif plot_type == 'points':
        p = ax.plot(xx, yy,  color=points_color, marker='.', transform=data_crs,
                    zorder=data_zorder)

    else:
        raise ValueError('plot_type  must be either "pcolormesh", "contourf", or "points"')

    gl = []
    cbar = []
    
    gl, cbar =\
        _add_features_to_axis(ax, p, cmin, cmax,
                              cmap = cmap,
                              show_coastline = show_coastline,
                              show_colorbar = show_colorbar,
                              show_land = show_land,
                              show_grid_lines = show_grid_lines,
                              show_grid_labels = show_grid_labels,
                              show_coastline_over_data = show_coastline_over_data,
                              show_land_over_data = show_land_over_data,
                              grid_linewidth = grid_linewidth,
                              grid_linestyle = grid_linestyle,
                              colorbar_label = colorbar_label,
                              colorbar_location = colorbar_location)

    return p, gl, cbar

# -----------------------------------------------------------------------------

def _add_features_to_axis(ax, p, cmin, cmax,
                          cmap = 'jet',
                          show_coastline=True,
                          show_colorbar=True,
                          show_land=True, 
                          show_grid_lines=True,
                          show_grid_labels=True,
                          show_coastline_over_data = True,
                          show_land_over_data = True,
                          grid_linewidth = 1,
                          grid_linestyle = '--',
                          colorbar_label = None,
                          colorbar_location = None):
                                     
    
    if show_land:
        if show_land_over_data:
        # place land over the data
            zorder = 75
        else:
        # place land under the data
            zorder = 25
        
        ax.add_feature(cfeature.LAND, zorder=zorder)
          
    if show_coastline:
        # place the coastline over land and over the data (default zorder 50)
        if show_coastline_over_data:
            zorder = 85
        else:
            # place coastline over land but under data
            zorder = 35
        
        ax.coastlines(linewidth=0.8, zorder=zorder)
        
    gl = []
    if show_grid_lines :
        # grid lines go over everything (zorder 110)
        gl = ax.gridlines(crs=ccrs.PlateCarree(),
                          linewidth=grid_linewidth, 
                          color='black',
                          draw_labels = show_grid_labels,
                          alpha=0.5, 
                          linestyle=grid_linestyle,
                          zorder=110)
        
    cbar = []
    if show_colorbar:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(cmin,cmax))
        sm._A = []
        if colorbar_location:
            cbar = plt.colorbar(sm,ax=ax, location=colorbar_location)
        else:
            cbar = plt.colorbar(sm,ax=ax)
            #cbar = plt.colorbar(p, ax=ax)
        if type(colorbar_label) is str:
            cbar.set_label(colorbar_label)
        
    return gl, cbar
# -----------------------------------------------------------------------------


def _create_projection_axis(projection_type,
                            user_lon_0,
                            user_lat_0,
                            parallels,
                            lat_lim,
                            subplot_grid,
                            less_output):

    """Set appropriate axis for projection type
    See plot_proj_to_latlon_grid for input parameter definitions.

    Returns
    -------
    ax :  matplotlib axis object
        defined with the correct projection
    show_grid_labels : logical
        True = show the grid labels, possibly not supported by all projections. 
               See cartopy documentation for updates.
    """

    if not less_output:
        print('_create_projection_axis: projection_type', projection_type)
        print('_create_projection_axis: user_lon_0, user_lat_0', user_lon_0, user_lat_0)
        print('_create_projection_axis: parallels', parallels)
        print('_create_projection_axis: lat_lim', lat_lim)

    # initialize (optional) subplot variables
    row = []
    col = []
    ind = []

    if subplot_grid is not None:

        if type(subplot_grid) is dict:
            row = subplot_grid['nrows']
            col = subplot_grid['ncols']
            ind = subplot_grid['index']

        elif type(subplot_grid) is list:
            row = subplot_grid[0]
            col = subplot_grid[1]
            ind = subplot_grid[2]

        else:
            raise TypeError('Unexpected subplot_grid type: ',type(subplot_grid))
    else:
        row = 1
        col = 1
        ind = 1

    proj_dict = dict()
    # Build dictionary of projection_types mapped to Cartopy calls
    # verify that each project is available in cartopy before adding it to the dictionary
    if hasattr(ccrs, 'EqualEarth'):
        proj_dict['EqualEarth']=ccrs.EqualEarth
    if hasattr(ccrs, 'Mercator'):
        proj_dict['Mercator']=ccrs.Mercator
    if hasattr(ccrs, 'LambertConformal'):
        proj_dict['LambertConformal']=ccrs.LambertConformal
    if hasattr(ccrs, 'AlbersEqualArea'):
        proj_dict['AlbersEqualArea']=ccrs.AlbersEqualArea
    if hasattr(ccrs, 'PlateCarree'):
        proj_dict['PlateCarree']=ccrs.PlateCarree
    if hasattr(ccrs, 'LambertCylindrical'):
        proj_dict['cyl']=ccrs.LambertCylindrical
    if hasattr(ccrs, 'Robinson'):
        proj_dict['robin']=ccrs.Robinson
    if hasattr(ccrs, 'Mollweide'):
        proj_dict['Mollweide']=ccrs.Mollweide
    if hasattr(ccrs, 'Orthographic'):
        proj_dict['ortho']=ccrs.Orthographic
    if hasattr(ccrs, 'InterruptedGoodeHomolosine'):
        proj_dict['InterruptedGoodeHomolosine']=ccrs.InterruptedGoodeHomolosine

    # stereo special cases
    if projection_type == 'stereo':
        if lat_lim>0:
            if hasattr(ccrs, 'NorthPolarStereo'):
               proj_dict['stereo']=ccrs.NorthPolarStereo
        else :
            if hasattr(ccrs, 'SouthPolarStereo'):
               proj_dict['stereo']=ccrs.SouthPolarStereo

    if projection_type not in proj_dict:
        raise NotImplementedError('requested projection type ', projection_type, ' is not an available projection:', proj_dict.keys())

    # Build dictionary for projection arguments
    proj_args={}
    if user_lon_0 is not None:
        proj_args['central_longitude']=user_lon_0
    if user_lat_0 is not None and projection_type != 'stereo':
        proj_args['central_latitude']=user_lat_0
    if (projection_type == 'LambertConformal') & (lat_lim is not None) :
        proj_args['cutoff']=lat_lim
    if parallels is not None :
        proj_args['standard_parallels']=parallels

    ax = plt.subplot(row, col, ind,
                    projection=proj_dict[projection_type](**proj_args))

    if not less_output:
        print('Projection type: ', projection_type)

    return ax



"""
Test routines for the vector calculations module
"""
from __future__ import division, print_function
import warnings
from pathlib import Path
import numpy as np
import xarray as xr
import pytest
from ecco_v4_py import scalar_calc, get_llc_grid

from .test_common import all_mds_datadirs, get_test_ds

def test_latitude_mask(get_test_ds):
    """run through lats, and ensure we're grabbing the closest
    "south-grid-cell-location" (whether that's in the x or y direction!)
    to each latitude value"""

    ds = get_test_ds

    grid = get_llc_grid(ds)
    wetC = ds['maskC'].isel(k=0)

    dLat = 0.5 # is this robust?
    nx = 90

    for lat in np.arange(-89,89,10):
        print('lat: ',lat)
        maskC = scalar_calc.get_latitude_mask(lat,ds['YC'],grid)

        maskC = maskC.where((maskC!=0) & wetC,0.)

        assert (ds['YC']-lat < dLat).where((maskC!=0)).all().values



"""
generic functions for scalar valued (cell centered) quantities
"""

import xarray as xr

def get_latitude_mask(lat_val,yc,grid):
    """Compute maskCedge which grabs the grid cell center points along 
    the desired latitude

    This mirrors the MATLAB function  gcmfaces/gcmfaces_calc/gcmfaces_lines_zonal.m

    Parameters
    ----------

    lat_val : int
        latitude at which to compute mask 
    yc : xarray DataArray
        Contains latitude values at cell centers
    grid : xgcm Grid object
        llc grid object generated via get_llc_grid

    Returns
    -------

    maskCedge : xarray DataArray
        contains mask of latitude at grid cell tracer points
    """

    # Compute difference in X, Y direction. 
    # multiply by 1 so that "True" -> 1, 2nd arg to "where" puts False -> 0 
    ones = xr.ones_like(yc)
    lat_maskC = ones.where(yc>=lat_val,0)

    maskCedge = get_edge_mask(lat_maskC,grid)

    return maskCedge

def get_edge_mask(maskC,grid):
    """From a given mask with points at cell centers, compute the 
    boundary between 1's and 0's

    Parameters
    ----------
    
    maskC : xarray DataArray
        containing 1's at interior points, 0's outside. We want the 
        boundary between them
    grid : xgcm Grid object

    Returns
    -------

    maskCedge : xarray DataArray
        with same dimensions as input maskC, with 1's at boundary 
        between 1's and 0's
    """

    # This first interpolation gets 0.5 at boundary points
    # however, the result lives on West and South grid cell edges
    maskX = grid.interp(maskC,'X', boundary='fill')
    maskY = grid.interp(maskC,'Y', boundary='fill')

    # Now interpolate these to get back on to cell centers
    # edge will now be at locations where values are 0.75
    maskXY= grid.interp_2d_vector({'X' : maskX, 'Y' : maskY}, boundary='fill')

    # Now wherever this is > 0 and the original mask is 0 is the boundary
    maskCedge = xr.ones_like(maskC).where( ((maskXY['X'] + maskXY['Y']) > 0) & (maskC==0.) , 0)

    return maskCedge



import warnings
from datetime import datetime
import numpy as np
import xarray as xr
import pytest
from ecco_v4_py import load_ecco_vars_from_mds

from .test_common import llc_mds_datadirs

@pytest.mark.parametrize("iters",['all',8,[0,8]])
@pytest.mark.parametrize("freq",['','AVG_MON','AVG_DAY','AVG_YEAR','SNAPSHOT'])
@pytest.mark.parametrize("vars_to_load",['all','U',['U','V','T']])
def test_read_mdsdataset(llc_mds_datadirs,iters,freq,vars_to_load):
    """honestly not sure what all of these options are for
    but just make sure this works"""

    data_dir,_ = llc_mds_datadirs
    ds = load_ecco_vars_from_mds(data_dir,
                                 mds_files=['U','V','W','T','S'],
                                 vars_to_load=vars_to_load,
                                 model_time_steps_to_load=iters,
                                 output_freq_code=freq)



"""
Module for computing meridional transport of quantities
"""

import numpy as np
import xarray as xr
# xarray compatibility
try:
    from xarray.core.pycompat import OrderedDict
except ImportError:
    from collections import OrderedDict

from .get_basin import get_basin_mask
from .ecco_utils import get_llc_grid
from ecco_v4_py import vector_calc

# Define constants
# These are chosen (for now) to match gcmfaces
METERS_CUBED_TO_SVERDRUPS = 10**-6
WATTS_TO_PETAWATTS = 10**-15
RHO_CONST = 1029
HEAT_CAPACITY = 4000

def calc_meridional_vol_trsp(ds,lat_vals,
                             basin_name=None,coords=None,grid=None):
    """Compute volumetric transport across latitude band in Sverdrups

    Parameters
    ----------
    ds : xarray Dataset
        must contain UVELMASS,VVELMASS, drF, dyG, dxG
    lat_vals : float or list
        latitude value(s) specifying where to compute transport
    basin_name : string, optional
        denote ocean basin over which to compute streamfunction
        If not specified, compute global quantity
        see get_basin.get_available_basin_names for options
    coords : xarray Dataset
        separate dataset containing the coordinate information
        YC, Z, drF, dyG, dxG, optionally maskW, maskS
    grid : xgcm Grid object, optional
        denotes LLC90 operations for xgcm, see ecco_utils.get_llc_grid
        see also the [xgcm documentation](https://xgcm.readthedocs.io/en/latest/grid_topology.html)


    Returns
    -------
    ds_out : xarray Dataset
        Dataset with the following variables
            'vol_trsp' :
                meridional volume transport in Sv
                with dimensions 'time' (if in given dataset) and 'lat'
            'vol_trsp_z' :
                meridional volume transport in Sv at each depth level
                in the given input fields
                dimensions: 'time' (if provided), 'lat', and 'k'
    """

    coords = _parse_coords(ds,coords,['Z','YC','drF','dyG','dxG'])

    x_vol = ds['UVELMASS'] * coords['drF'] * coords['dyG']
    y_vol = ds['VVELMASS'] * coords['drF'] * coords['dxG']

    # Computes salt transport in m^3/s at each depth level
    ds_out = meridional_trsp_at_depth(x_vol,y_vol,
                                      coords=coords,
                                      lat_vals=lat_vals,
                                      basin_name=basin_name,
                                      grid=grid)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'vol_trsp_z'})

    # Sum over depth for total transport
    ds_out['vol_trsp'] = ds_out['vol_trsp_z'].sum('k')

    # Convert both fields to Sv
    for fld in ['vol_trsp','vol_trsp_z']:
        ds_out[fld] = METERS_CUBED_TO_SVERDRUPS * ds_out[fld]
        ds_out[fld].attrs['units'] = 'Sv'

    return ds_out


def calc_meridional_heat_trsp(ds,lat_vals,
                              basin_name=None,coords=None,grid=None):
    """Compute heat transport across latitude band in Petwatts
    see calc_meridional_vol_trsp for argument documentation.
    The only differences are:

    Parameters
    ----------
    ds : xarray Dataset
        must contain fields 'ADVx_TH','ADVy_TH','DFxE_TH','DFyE_TH'
    coords : xarray Dataset, optional
        in case coordinates are in a separate dataset
        only needs field 'YC' and optionally maskW, maskS

    Returns
    -------
    ds_out : xarray Dataset
        Dataset with the following variables
            'heat_trsp' :
                meridional heat transport in PW
                with dimensions 'time' (if in given dataset) and 'lat'
            'heat_trsp_z' :
                meridional heat transport in PW at each depth level
                in the given input fields
                dimensions: 'time' (if provided), 'lat', and 'k'
    """

    coords = _parse_coords(ds,coords,['Z','YC'])

    x_heat = ds['ADVx_TH'] + ds['DFxE_TH']
    y_heat = ds['ADVy_TH'] + ds['DFyE_TH']

    # Computes heat transport in degC * m^3/s at each depth level
    ds_out = meridional_trsp_at_depth(x_heat,y_heat,
                                      coords=coords,
                                      lat_vals=lat_vals,
                                      basin_name=basin_name,
                                      grid=grid)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'heat_trsp_z'})

    # Sum over depth for total transport
    ds_out['heat_trsp'] = ds_out['heat_trsp_z'].sum('k')

    # Convert both fields to PW
    for fld in ['heat_trsp','heat_trsp_z']:
        ds_out[fld] = WATTS_TO_PETAWATTS * RHO_CONST * HEAT_CAPACITY * ds_out[fld]
        ds_out[fld].attrs['units'] = 'PW'

    return ds_out

def calc_meridional_salt_trsp(ds,lat_vals,
                              basin_name=None,coords=None,grid=None):
    """Compute salt transport across latitude band in psu * Sv
    see calc_meridional_vol_trsp for argument documentation.
    The only differences are:

    Parameters
    ----------
    ds : xarray Dataset
        must contain fields 'ADVx_SLT','ADVy_SLT','DFxE_SLT','DFyE_SLT'
    coords : xarray Dataset, optional
        in case coordinates are in a separate dataset
        only needs field 'YC' and optionally maskW, maskS

    Returns
    -------
    ds_out : xarray Dataset
        Dataset with the following variables
            'salt_trsp' :
                meridional salt transport in psu*Sv
                with dimensions 'time' (if in given dataset) and 'lat'
            'salt_trsp_z' :
                meridional salt transport in psu*Sv at each depth level
                in the given input fields
                dimensions: 'time' (if provided), 'lat', and 'k'
    """

    coords = _parse_coords(ds,coords,['Z','YC'])

    x_salt = ds['ADVx_SLT'] + ds['DFxE_SLT']
    y_salt = ds['ADVy_SLT'] + ds['DFyE_SLT']

    # Computes salt transport in psu * m^3/s at each depth level
    ds_out = meridional_trsp_at_depth(x_salt,y_salt,
                                      coords=coords,
                                      lat_vals=lat_vals,
                                      basin_name=basin_name,
                                      grid=grid)

    # Rename to useful data array name
    ds_out = ds_out.rename({'trsp_z': 'salt_trsp_z'})

    # Sum over depth for total transport
    ds_out['salt_trsp'] = ds_out['salt_trsp_z'].sum('k')

    # Convert both fields to psu.Sv
    for fld in ['salt_trsp','salt_trsp_z']:
        ds_out[fld] = METERS_CUBED_TO_SVERDRUPS * ds_out[fld]
        ds_out[fld].attrs['units'] = 'psu.Sv'

    return ds_out

# ---------------------------------------------------------------------

def meridional_trsp_at_depth(xfld, yfld, lat_vals, coords,
                             basin_name=None, grid=None, less_output=True):
    """
    Compute transport of vector quantity at each depth level
    across latitude(s) defined in lat_vals

    Parameters
    ----------
    xfld, yfld : xarray DataArray
        3D spatial (+ time, optional) field at west and south grid cell edges
    lat_vals : float or list
        latitude value(s) specifying where to compute transport
    coords : xarray Dataset
        only needs YC, and optionally maskW, maskS (defining wet points)
    basin_name : string, optional
        denote ocean basin over which to compute streamfunction
        If not specified, compute global quantity
        see get_basin.get_available_basin_names for options
    grid : xgcm Grid object, optional
        denotes LLC90 operations for xgcm, see ecco_utils.get_llc_grid
        see also the [xgcm documentation](https://xgcm.readthedocs.io/en/latest/grid_topology.html)

    Returns
    -------
    ds_out : xarray Dataset
        with the main variable
            'trsp_z'
                transport of vector quantity across denoted latitude band at
                each depth level with dimensions 'time' (if in given dataset),
                'k' (depth), and 'lat'
    """

    if grid is None:
        grid = get_llc_grid(coords)

    # Initialize empty DataArray with coordinates and dims
    ds_out = _initialize_trsp_data_array(coords, lat_vals)

    # Get basin mask
    maskW = coords['maskW'] if 'maskW' in coords else xr.ones_like(xfld)
    maskS = coords['maskS'] if 'maskS' in coords else xr.ones_like(yfld)
    if basin_name is not None:
        maskW = get_basin_mask(basin_name,maskW)
        maskS = get_basin_mask(basin_name,maskS)

    # These sums are the same for all lats, therefore precompute to save
    # time
    tmp_x = xfld.where(maskW)
    tmp_y = yfld.where(maskS)

    for lat in ds_out.lat.values:
        if not less_output:
            print ('calculating transport for latitutde ', lat)

        # Compute mask for particular latitude band
        lat_maskW, lat_maskS = vector_calc.get_latitude_masks(lat, coords['YC'], grid)

        # Sum horizontally
        lat_trsp_x = (tmp_x * lat_maskW).sum(dim=['i_g','j','tile'])
        lat_trsp_y = (tmp_y * lat_maskS).sum(dim=['i','j_g','tile'])

        ds_out['trsp_z'].loc[{'lat':lat}] = lat_trsp_x + lat_trsp_y

    return ds_out


def _initialize_trsp_data_array(coords, lat_vals):
    """Create an xarray DataArray with time, depth, and latitude dims

    Parameters
    ----------
    coords : xarray Dataset
        contains LLC coordinates 'k' and (optionally) 'time'
    lat_vals : int or array of ints
        latitude value(s) rounded to the nearest degree
        specifying where to compute transport

    Returns
    -------
    ds_out : xarray Dataset
        Dataset with the variables
            'trsp_z'
                zero-valued DataArray with time (optional),
                depth, and latitude dimensions
            'Z'
                the original depth coordinate
    """

    lat_vals = np.array(lat_vals) if isinstance(lat_vals,list) else lat_vals
    lat_vals = np.array([lat_vals]) if np.isscalar(lat_vals) else lat_vals
    lat_vals = xr.DataArray(lat_vals,coords={'lat':lat_vals},dims=('lat',))

    xda = xr.zeros_like(coords['k']*lat_vals)
    xda = xda if 'time' not in coords.dims else xda.broadcast_like(coords['time']).copy()

    # Convert to dataset to add Z coordinate
    xds = xda.to_dataset(name='trsp_z')
    xds['Z'] = coords['Z']
    xds = xds.set_coords('Z')

    return xds

def _parse_coords(ds,coords,coordlist):
    if coords is not None:
        return coords
    else:
        for f in set(['maskW','maskS']).intersection(ds.reset_coords().keys()):
            coordlist.append(f)

        if 'time' in ds.dims:
            coordlist.append('time')

        dsout = ds[coordlist]
        if 'domain' in ds.attrs:
            dsout.attrs['domain'] = ds.attrs['domain']
        return dsout



"""
ECCO v4 Python: tile_io

This module provides routines for loading ECCO netcdf files.
--- now with actual documentation!

.. _ecco_v4_py Documentation :
   https://github.com/ECCO-GROUP/ECCOv4-py

"""
from __future__ import division,print_function
import numpy as np
import xarray as xr
import glob
import os
import re
from pathlib import Path
from collections import OrderedDict


#%%
def load_ecco_grid_nc(grid_dir, grid_filename, \
                      tiles_to_load = 'all', \
                      k_subset = [], \
                      dask_chunk = False,\
                      less_output = True):

    """

    Loads the ECCOv4 NetCDF model grid parameters.
    All 13 tiles of the lat-lon-cap (llc) grid are present in this
    type of file.  A subset of the vertical level (k_subset) and the
    model grid tiles (tiles_to_load) are optional arguments.

    Simply pass the directory with the model grid file

    Parameters
    ----------
    grid_dir : str
        path to a directory within which we will look for NetCDF grid file

    grid_filename : str
        name of model grid file
        filename should be something like : ECCOv4r3_grid.nc or ECCOv4r4_grid.nc

    tiles_to_load : int or list or range, optional, default range(13)
        a list of which tiles to load.

    k_subset : list, optional, default = [] (load all)
        a list of which vertical levels to load.

    dask_chunk : boolean, optional, default True
        whether or not to ask Dask to chunk the arrays into pieces 90x90.
        WARNING: DON'T MESS WITH THIS NUMBER.

    less_output : boolean, default False
        A debugging flag.  False = less debugging output

    Returns
    -------
    g : Dataset
        an xarray Dataset

    """
    if not less_output:
        print ('tiles_to_load ', tiles_to_load)

    if isinstance(tiles_to_load,str):
        if 'all' in tiles_to_load:
            tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12]
        else:
            tiles_to_load = list(int(tiles_to_load))
    elif isinstance(tiles_to_load,tuple):
        tiles_to_load = list(tiles_to_load)
    elif isinstance(tiles_to_load, int):
        tiles_to_load = [tiles_to_load]
    elif isinstance(tiles_to_load, range):
        tiles_to_load = list(tiles_to_load)
    elif not isinstance(tiles_to_load, list):
        raise Exception ('tiles_to_load has to be a tuple, int, list or string ' + \
                         'you passed a %s and I cannot handle that' % type(tiles_to_load))

    if not less_output:
        print(grid_filename, grid_dir)

    file = Path(grid_dir) / grid_filename

    if file.exists():

        if not less_output:
            print ('--- LOADING model grid file: %s' % file)

        if dask_chunk:
            g_i = xr.open_dataset(str(file), chunks=90)
        else:
            g_i = xr.open_dataset(str(file)).load()

        # pull out a k subset
        if len(k_subset) > 0 :
            if 'k' in g_i.coords.keys():
                g_i = g_i.isel(k=k_subset)

            if 'k_u' in g_i.coords.keys():
                g_i = g_i.isel(k_u=k_subset)

            if 'k_l' in g_i.coords.keys():
                g_i = g_i.isel(k_l=k_subset)

            if 'k_p1' in g_i.coords.keys():
                g_i = g_i.isel(k_p1=k_subset)


        # pull out the tile subset
        g_i = g_i.isel(tile=tiles_to_load)


        g_i.attrs = OrderedDict(sorted(g_i.attrs.items()))

        return g_i

    # no files
    else:
        print('\n\n Attention!!')
        print('ECCO netcdf grid file not found in ' + str(file))
        return []


#%%
def load_ecco_var_from_years_nc(data_dir, var_to_load, \
                                years_to_load = 'all', \
                                tiles_to_load = 'all', \
                                k_subset = [], \
                                dask_chunk = True,\
                                less_output = True):

    """

    Loads one or more ECCOv4 NetCDF state estimate variable in the
    format of one file per variable per year.  All 13 tiles of the
    lat-lon-cap (llc) grid are present in this type of file.

    Files in this format have names like
        /eccov4r3_native_grid_netcdf/mon_mean/THETA/THETA_2010.nc
        /eccov4r3_native_grid_netcdf/day_mean/THETA/THETA_2010.nc
        /eccov4r3_native_grid_netcdf/mon_snapshot/THETA/THETA_2010.nc

    Simply point this routine at a directory with one or more
    of these files and one or more years of var_to_load variables will
    be loaded.
    * Used repeatedly by recursive_load_ecco_var_from_years_nc *


    Parameters
    ----------
    data_dir : str
        path to a directory within which we will look for NetCDF tile files
        something like: '/eccov4r3_native_grid_netcdf/mon_mean/THETA/'

    tiles_to_load : int or list or range, optional, default range(13)
        a list of which tiles to load.

    var_to_load : str
        string indicating which variable to load.

    years_to_load : str, int or list, optional, default 'all'
        a list of which years to load.

    k_subset : list, optional, default = [] (load all)
        a list of which vertical levels to load.

    dask_chunk : boolean, optional, default True
        whether or not to ask Dask to chunk the arrays into pieces 90x90.
        WARNING: DON'T MESS WITH THIS NUMBER.

    less_output : boolean, default False
        A debugging flag.  False = less debugging output

    Returns
    -------
    g : Dataset
        an xarray Dataset

    """
    if not less_output:
        print ('tiles_to_load ', tiles_to_load)
        print ('years to load ', years_to_load)

    if isinstance(tiles_to_load,str):
        if 'all' in tiles_to_load:
            tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12]
        else:
            tiles_to_load = list(int(tiles_to_load))
    elif isinstance(tiles_to_load,tuple):
        tiles_to_load = list(tiles_to_load)
    elif isinstance(tiles_to_load, int):
        tiles_to_load = [tiles_to_load]
    elif isinstance(tiles_to_load, range):
        tiles_to_load = list(tiles_to_load)
    elif not isinstance(tiles_to_load, list):
        raise Exception ('tiles_to_load has to be a tuple, int, list or string ' + \
                         'you passed a %s and I cannot handle that' % type(tiles_to_load))


    var_path = Path(data_dir)
    files = sorted(list(var_path.glob('**/*nc')))

    if not less_output:
        print ('--- LOADING %s FROM YEARS NC: %s' % \
               (var_to_load, var_path))
        print (files)

    if isinstance(years_to_load, str):
        if 'all' not in years_to_load:
            years_to_load = int(years_to_load)
    elif isinstance(years_to_load,tuple):
        years_to_load = list(years_to_load)
    elif isinstance(years_to_load, int):
        years_to_load = [years_to_load]
    elif isinstance(years_to_load, str):
        years_to_load = int(years_to_load)
    elif isinstance(years_to_load, range):
        years_to_load = list(years_to_load)
    elif not isinstance(years_to_load, list):
        raise Exception ('years_to_load has to be a tuple, int, list or string ' + \
                         'you passed a %s and I cannot handle that' % type(years_to_load))

    # g will be the DataArray for this variable and all times
    g = []
    if len(files) > 0:
        if not less_output:
            print ('---found %s nc files here.  loading ....' % var_to_load)

        for file in files:
            file_basename =  file.stem
            if not less_output:
                print('file basename : ', file_basename)

            split_list = file.stem.split(sep='_')
            file_year_found = False
            var_name_of_file = []
            file_year = []
            for pi, piece in enumerate(split_list):
                if (is_year(piece)) and (file_year_found == False):
                    file_year = int(piece)
                    var_name_of_file = file.stem.split(sep='_' + piece)[0]
                    file_year_found = True
                    if not less_output:
                        print (var_name_of_file, file_year)

            if var_name_of_file == var_to_load:
                if not less_output:
                    print('var name matches ', var_name_of_file)

                if 'all' in years_to_load or file_year in years_to_load:
                    if not less_output:
                        print ('year in years_to_load', file_year)

                    if dask_chunk:
                        g_i = xr.open_dataset(str(file), chunks=90)
                    else:
                        g_i = xr.open_dataset(str(file)).load()

                    # pull out a k subset
                    if len(k_subset) > 0 and 'k' in g_i.coords.keys():
                        g_i = g_i.isel(k=k_subset)

                    # pull out the tile subset
                    g_i = g_i.isel(tile=tiles_to_load)


                    if isinstance(g, list):
                        g = g_i
                    else:
                        g = xr.concat((g, g_i),'time')


        # finished looping through files
        if len(g) == 0:
            if not less_output:
                print ('we had files but did not load any matching %s ' % \
                       var_to_load)

        else:
            g.attrs = OrderedDict(sorted(g.attrs.items()))

    # no files
    else:
        if not less_output:
            print ('no files found with name "%s" in %s \n' % \
                   (var_to_load, data_dir ))

    return g

#%%
def is_year(s):
    try:
        return int(s) >= 1900
    except ValueError:
        return False

#%%
def recursive_load_ecco_var_from_years_nc(data_root_dir,
                                          vars_to_load = 'all',
                                          tiles_to_load = 'all',
                                          years_to_load = 'all',
                                          k_subset = [],
                                          dask_chunk = True,
                                          less_output = True):

    """

    Loads one or more state estimate variables for one or more
    years.  Appropriate for ECCOv4 NetCDF files stored in the
    format of one file per variable per year.  All 13 tiles of the
    lat-lon-cap (llc) grid are present in this type of file.

    Files in this format have names like
        /eccov4r3_native_grid_netcdf/mon_mean/THETA_2010.nc
        /eccov4r3_native_grid_netcdf/day_mean/THETA_2010.nc
        /eccov4r3_native_grid_netcdf/mon_snapshot/THETA_2010.nc

    Simply point this routine at a top-level directory, a list of
    variables you want, and a list of years and prepare to be dazzled.

    * Makes heavy use of load_ecco_var_from_years_nc *


    Parameters
    ----------
    data_root_dir : str
        path to a top-level directory below we will look for NetCDF tile files

    vars_to_load : list or str, optional, default 'all'
        a list or string indicating which variables you want to load.

        Note: if 'all', data_root_dir must be a directory with one or more
        variable names. In the follow example, THETA, SALT, and ETAN
        will be loaded if the full path to `eccov4_native_grid_netcdf` is
        provided:
            /eccov4r3_native_grid_netcdf/THETA/THETA_YYYY.nc
            /eccov4r3_native_grid_netcdf/SALT/SALT_YYYY.nc
            /eccov4r3_native_grid_netcdf/ETAN/ETAN_YYYY.nc

    tiles_to_load : int or list or range, optional, default range(13)
        a list of which tiles to load.

    years_to_load : int or list or range, optional, default 'all'
        a list of which tiles to load

    k_subset : list, optional, default = [] (load all)
        a list of which vertical levels to load.

    dask_chunk : boolean, optional, default True
        whether or not to ask Dask to chunk the arrays into pieces 90x90.
        WARNING: DON'T MESS WITH THIS NUMBER.

    less_output : boolean, default False
        A debugging flag.  False = less debugging output

    Returns
    -------
    g : Dataset
        an xarray Dataset

    """

    if not less_output:
        print ('tiles_to_load ', tiles_to_load)
        print ('years to load ', years_to_load)

    if isinstance(tiles_to_load,str):
        if 'all' in tiles_to_load:
            tiles_to_load = [0,1,2,3,4,5,6,7,8,9,10,11,12]
        else:
            tiles_to_load = list(int(tiles_to_load))
    elif isinstance(tiles_to_load,tuple):
        tiles_to_load = list(tiles_to_load)
    elif isinstance(tiles_to_load, int):
        tiles_to_load = [tiles_to_load]
    elif isinstance(tiles_to_load, range):
        tiles_to_load = list(tiles_to_load)
    elif not isinstance(tiles_to_load, list):
        raise Exception ('tiles_to_load has to be a tuple, int, list or string ' + \
                         'you passed a %s and I cannot handle that' % str(type(tiles_to_load)))

    if isinstance(years_to_load, str):
        if 'all' not in years_to_load:
            years_to_load = int(years_to_load)
    elif isinstance(years_to_load,tuple):
        years_to_load = list(years_to_load)
    elif isinstance(years_to_load, int):
        years_to_load = [years_to_load]
    elif isinstance(years_to_load, str):
        years_to_load = int(years_to_load)
    elif isinstance(years_to_load, range):
        years_to_load = list(years_to_load)
    elif not isinstance(years_to_load, list):
        raise Exception ('years_to_load has to be a tuple, int, list or string ' + \
                         'you passed a %s and I cannot handle that' % type(years_to_load))

    if not isinstance(vars_to_load, list):
        vars_to_load = [vars_to_load]


    var_path = Path(data_root_dir)

    files = sorted(list(var_path.glob('**/*nc')))

    all_var_names = []

    files_to_load_dict = dict()

    if len(files) == 0:
        print ('no netcdf files found in this directory or subdirectories')
        return []

    # loop through all netcdf files found in subdirectories of 'data_root_dir'
    for file in files:
        if not less_output:
            print('file basename : ', file.stem)

        # find the variable name and year.
        # assumes format is VARNAME_YYYY_other stuff.nc
        # we have to deal with the fact that some variables have names with '_' in them
        # so first we look for the YYYY and then assume everything to the left of that
        # is the variable name.  e.g., ADVx_TH_1999 =>  ADVx_TH, 1999
        split_list = file.stem.split(sep='_')
        file_year_found = False
        var_name_of_file = []
        file_year = []
        for pi, piece in enumerate(split_list):
            if (is_year(piece)) and (file_year_found == False):
                file_year = int(piece)
                var_name_of_file = file.stem.split(sep='_' + piece)[0]
                file_year_found = True
                if not less_output:
                    print (var_name_of_file, file_year)

        if not var_name_of_file in all_var_names:
            all_var_names.append(var_name_of_file)

        if 'all' in vars_to_load or var_name_of_file in vars_to_load:
            if 'all' in years_to_load or file_year in years_to_load:

                # add this file to a new list of files to read
                if var_name_of_file not in files_to_load_dict.keys():
                    files_to_load_dict[var_name_of_file] = [file]
                else:
                    # add this file to the existing list of files to read
                    files_to_load_dict[var_name_of_file].append(file)

    # g_all will be a dataset containing all data arrays
    g_all = []

    # loop through all variables that we are going to load
    for var_to_load in files_to_load_dict.keys():

        # g is a data array that will contain all time levels
        # for this variable
        g = []
        print('loading files of ', var_to_load)

        # get all filenames corresponding to this variable to load
        files_for_var = files_to_load_dict[var_to_load]

        # loop through all of those variables
        for file in files_for_var:

            if not less_output:
                print('loading ', str(file))

            # g_i is a data arary for one time level of this variable
            if dask_chunk:
                g_i = xr.open_dataset(str(file), chunks=90)
            else:
                g_i = xr.open_dataset(str(file)).load()

            # pull out a k subset
            if len(k_subset) > 0 and 'k' in g_i.coords.keys():
                g_i = g_i.isel(k=k_subset)

            # pull out the tile subset
            g_i = g_i.isel(tile=tiles_to_load)

            # combine this time level (g_i) with the data array that contains
            # all time levels (g)
            if isinstance(g, list):
                g = g_i
            else:
                g = xr.concat((g, g_i),'time')


        # finished looping through files, check to make sure we have something
        # in g
        if len(g) == 0:
            if not less_output:
                print ('we had files but did not load any matching %s ' % \
                       var_to_load)

        else:

            # Add this DataArray to the DataSet g_all
            # -- if this is the first DataArray, then make g_all, g
            if len(g_all) == 0:
                g_all = g

            # otherwise merge g into g_all
            else:
                g_all = xr.merge((g_all,g))

                # when you merge DataArrays or DataSets some metadata can
                # be lost.  This bit ensures that all the metadata is carried
                # over to g_all
                for g_attr_key in g.attrs.keys():
                    if g_attr_key not in g_all.attrs.keys():
                        g_all.attrs[g_attr_key] = g.attrs[g_attr_key]

    # finished loading all variable DataArrays
    # update metadata again
    if len(g_all) > 0:
        g_all.attrs = OrderedDict(sorted(g_all.attrs.items()))
    else:
        print('we ended up with nothing loaded!')

    return g_all




"""
Functions defined on vector valued fields
"""

import xarray as xr
import xgcm
from .ecco_utils import get_llc_grid


def UEVNfromUXVY(x_fld,y_fld, coords, grid=None):
    """Compute the zonal and meridional components of a vector field defined
    by its x and y components with respect to the model grid.

    The x and y components of the vector can be defined on model grid cell edges ('u' and 'v' points),
    or on model grid cell centers ('c' points). If the vector components are defined on the grid cell edges then the function will first interpolate them to the grid cell centers. 
    
    Once both x and y vector components are at the cell centers, they are rotated to the zonal and meridional components using the cosine and sine of the grid cell angle. The grid cell angle is defined as the angle between the Earth's parallels and the line connecting the center of a grid cell to the center of its neighbor in the x-direction. The cosine and sine of this angle are provided in the 'coords' input field.

    The function will raise an error if the grid cell angle terms (CS, SN) are not provided in the coords

    Example vector fields provided at the grid cell edges are UVEL and VVEL. Example fields provided at the grid cell centers are EXFuwind and EXFvwind. 

    Note: this routine is inspired by gcmfaces_calc/calc_UEVNfromUXVY.m

    Parameters
    ----------
    x_fld, y_fld : xarray DataArray
        x and y components of a vector field provided at the model grid cell edges or centers
    coords : xarray Dataset
        must contain CS (cosine of grid orientation) and SN (sine of grid orientation)
    grid : xgcm Grid object, optional
        see ecco_utils.get_llc_grid and xgcm.Grid
        If not provided, the function will create one using the coords dataset.
    
    Returns
    -------
    e_fld, n_fld : xarray DataArray
        zonal (positive east) and meridional (positive north) components of input vector field 
        at grid cell centers/tracer points
    """

    # Check to make sure 'CS' and 'SN' are in coords
    # before doing calculation
    required_fields = ['CS','SN']
    for var in required_fields:
        if var not in coords.variables:
            raise KeyError('Could not find %s in coords Dataset' % var)

    # If no grid, establish it
    if grid is None:
        grid = get_llc_grid(coords)

    # Determine if vector fields are at cell edges or cell centers
    # by checking the dimensions of the x and y fields
    # If the vector components are at cell edges, we need to interpolate to cell centers
    # If the vector components are at cell centers, we don't need to interpolate
    # Create a set with all of the dimensions of the x and y fields
    vector_dims = set(x_fld.dims + y_fld.dims)
    
    # if neither 'i_g' and 'j_g' are present then the vector components are at cell centers
    vector_components_at_cell_center = 'i_g' not in vector_dims and 'j_g' not in vector_dims

    if vector_components_at_cell_center:
        # If the vector components are at cell centers, we don't need to interpolate
        # vec_field is a dictionary with the x and y fields
        vec_field = {'X': x_fld, 'Y': y_fld}
    else:
        # If the vector components are at cell edges, we need to interpolate to cell centers
        # vec_field is a dictionary with the x and y fields
        vec_field = grid.interp_2d_vector({'X': x_fld, 'Y': y_fld},boundary='fill')
        
    # Compute the zonal "e" and meridional components "n" using cos(), sin()
    e_fld = vec_field['X']*coords['CS'] - vec_field['Y']*coords['SN']
    n_fld = vec_field['X']*coords['SN'] + vec_field['Y']*coords['CS']

    return e_fld, n_fld



def get_latitude_masks(lat_val,yc,grid):
    """Compute maskW/S which grabs vector field grid cells along specified latitude
    band and corrects the sign associated with X-Y LLC grid

    This mirrors the MATLAB function gcmfaces/gcmfaces_calc/gcmfaces_lines_zonal.m

    Parameters
    ----------

    lat_val : int
        latitude at which to compute mask
    yc : xarray DataArray
        Contains latitude values at cell centers
    grid : xgcm Grid object
        llc grid object generated via get_llc_grid

    Returns
    -------

    maskWedge, maskSedge : xarray DataArray
        contains masks of latitude band at grid cell west and south grid edges
    """

    # Compute difference in X, Y direction.
    # multiply by 1 so that "True" -> 1, 2nd arg to "where" puts False -> 0
    ones = xr.ones_like(yc)
    maskC = ones.where(yc>=lat_val,0)

    maskWedge = grid.diff( maskC, 'X', boundary='fill')
    maskSedge = grid.diff( maskC, 'Y', boundary='fill')

    return maskWedge, maskSedge

